#    Copyright 2021 Eugene Frolov.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import abc
import collections
import threading

from restalchemy.storage import base
from restalchemy.storage.sql.dialect.query_builder import common
from restalchemy.storage.sql import filters as sql_filters


class Table(common.AbstractClause):
    def __init__(self, model, session):
        super(Table, self).__init__(session)
        self._name = model.__tablename__
        self._model = model
        self._columns = self._build_columns(self._model)

    @property
    def name(self):
        return self._name

    @property
    def model(self):
        return self._model

    def _build_columns(self, model):
        # Note(efrolov): to save ordering
        ordered_result = collections.OrderedDict()
        for name, prop in model.properties.properties.items():
            ordered_result[name] = common.Column(name, prop, self._session)
        return ordered_result

    def get_columns(self, with_prefetch=True):
        return [
            column
            for column in self._columns.values()
            if not column.model_property.is_prefetch() or with_prefetch
        ]

    def get_prefetch_columns(self):
        return [
            column
            for column in self._columns.values()
            if column.model_property.is_prefetch()
        ]

    def get_column_by_name(self, name):
        return self._columns[name]

    def compile(self):
        return self._session.engine.escape(self._name)


class Limit(common.AbstractClause):
    def __init__(self, value, session):
        super(Limit, self).__init__(session)
        self._value = value

    def compile(self):
        return "LIMIT %d" % self._value


class For(common.AbstractClause):
    def __init__(self, session, share=False):
        super(For, self).__init__(session)
        self._is_share = share

    def compile(self):
        return "FOR %s" % ("SHARE" if self._is_share else "UPDATE")


class Criteria(common.AbstractClause, metaclass=abc.ABCMeta):
    def __init__(self, clause1, clause2, session):
        super(Criteria, self).__init__(session)
        self._clause1 = clause1
        self._clause2 = clause2


class EQCriteria(Criteria):
    def compile(self):
        return "%s = %s" % (
            self._clause1.original.compile(),
            self._clause2.original.compile(),
        )


class On(common.AbstractClause):
    def __init__(self, list_of_criteria, session):
        super(On, self).__init__(session)
        self._list_of_criteria = list_of_criteria

    def compile(self):
        return " AND ".join([c.compile() for c in self._list_of_criteria])


class LeftJoin(common.AbstractClause):
    def __init__(self, table, on, session):
        # type: (common.TableAlias, On) -> LeftJoin
        super(LeftJoin, self).__init__(session)
        self._table = table
        self._on = on

    def compile(self):
        return "LEFT JOIN %s ON (%s)" % (
            self._table.compile(),
            self._on.compile(),
        )


class OrderByValue(common.AbstractClause):
    SORT_TYPES = frozenset(
        (
            "ASC",
            "ASC NULLS FIRST",
            "ASC NULLS LAST",
            "DESC",
            "DESC NULLS FIRST",
            "DESC NULLS LAST",
        )
    )

    def __init__(self, column, session, sort_type=None):
        super(OrderByValue, self).__init__(session)
        self._column = column
        if not sort_type:
            self._sort_type = "ASC"
        else:
            self._sort_type = sort_type.upper()
            if self._sort_type not in self.SORT_TYPES:
                raise ValueError("Unknown order: %s" % self._sort_type)

    def compile(self):
        """
        Generic compilation of the ORDER BY clause.

        A note regarding "NULLS FIRST/LAST":
            Resulting SQL looks like:
                ORDER BY
                    CASE WHEN table.col IS NULL THEN 0 ELSE 1 END ASC,
                    table.col ASC

            So, for example with NULLS FIRST:
            CASE result | Actual value
            ----------- | ------------
             0          | NULL
             0          | NULL
             1          | 2
             1          | 5
             1          | 8

            This approach is generic, and works for both PostgreSQL and MySQL,
            unlike PostgreSQL-specific "NULLS FIRST/LAST".

            todo: Add PostgreSQL-specific ordering using the native
                NULLS FIRST/LAST syntax for its performance benefits.
        """
        column_name = self._column.compile()

        # Handle simple ASC/DESC without NULLS specification
        if self._sort_type in ("ASC", "DESC"):
            return "%s %s" % (column_name, self._sort_type)

        # Handle NULLS FIRST/LAST with a generic CASE approach
        order_map = {
            "ASC NULLS FIRST": ("ASC", 0, 1),
            "ASC NULLS LAST": ("ASC", 1, 0),
            "DESC NULLS FIRST": ("DESC", 0, 1),
            "DESC NULLS LAST": ("DESC", 1, 0),
        }
        order, then_, else_ = order_map[self._sort_type]
        return (
            f"CASE WHEN {column_name} IS NULL "
            f"THEN {then_} ELSE {else_} END ASC, "
            f"{column_name} {order}"
        )


class ResultField(object):
    def __init__(self, alias_name):
        super(ResultField, self).__init__()
        self._alias_name = alias_name

    def parse(self, row):
        return row[self._alias_name]


class ResultNode(object):
    def __init__(self):
        super(ResultNode, self).__init__()
        self._child_nodes = {}

    def add_child_field(self, name, alias_name):
        self._child_nodes[name] = ResultField(alias_name=alias_name)
        return self._child_nodes[name]

    def add_child_node(self, name):
        self._child_nodes[name] = ResultNode()
        return self._child_nodes[name]

    def parse(self, row):
        result = base.PrefetchResult()
        for name, child_node in self._child_nodes.items():
            result[name] = child_node.parse(row)
        return result


class ResultParser(object):
    def __init__(self):
        super(ResultParser, self).__init__()
        self._root = ResultNode()

    @property
    def root(self):
        return self._root


class SelectQ(common.AbstractClause):
    def __init__(self, model, session):
        super(SelectQ, self).__init__(session)
        self._autoinc = 0
        self._autoinc_lock = threading.RLock()
        self._result_parser = ResultParser()
        self._model_table = common.TableAlias(
            Table(model, session=session),
            self._build_table_alias_name(),
            session=session,
        )
        self._select_expressions = []
        self._table_references = [self._model_table]  # type: list
        self._where_expression = sql_filters.AND()
        self._order_by_expressions = []
        self._for_expression = None
        self._limit_condition = None
        self._add_column_to_select_expressions(
            result_parser_node=self._result_parser.root,
            columns=self._model_table.get_columns(with_prefetch=False),
        )
        self._resolve_model_dependency(
            table=self._model_table,
            result_parser_node=self._result_parser.root,
        )

    def _resolve_model_dependency(self, table, result_parser_node):
        for column in table.get_prefetch_columns():
            dep_model = column.model_property.get_property_type()

            # Search primary key column
            id_properties = dep_model.get_id_property()
            if len(id_properties) != 1:
                msg = (
                    "Can't automatic resolve dependency for %s table"
                    " because the number of fields for primary keys (%r)"
                    " of model (%r) is not equal to 1."
                ) % (table.name, id_properties, dep_model)
                raise ValueError(msg)
            alias = common.TableAlias(
                Table(dep_model, session=self._session),
                self._build_table_alias_name(),
                session=self._session,
            )
            id_column = alias.get_column_by_name(list(id_properties.keys())[0])

            # Construct Left Join for prefetch dependency
            left_join = LeftJoin(
                table=alias,
                on=On(
                    [EQCriteria(column, id_column, session=self._session)],
                    session=self._session,
                ),
                session=self._session,
            )
            self._table_references.append(left_join)

            # Adding columns to fetch data on it
            node = result_parser_node.add_child_node(column.original_name)
            self._add_column_to_select_expressions(
                result_parser_node=node,
                columns=alias.get_columns(with_prefetch=False),
            )

            # Processing parent model to resolve dependencies
            self._resolve_model_dependency(
                table=alias,
                result_parser_node=node,
            )

    def _add_column_to_select_expressions(self, result_parser_node, columns):
        for column in columns:
            result_parser_node.add_child_field(
                column.original_name,
                column.name,
            )
            self._select_expressions.append(column)
        return self._select_expressions

    def _wrap_alias(self, table, fields):
        return [
            common.ColumnAlias(
                field,
                "%s_%s" % (table.name, field.name),
                session=self._session,
            )
            for field in fields
        ]

    def where(self, filters=None):
        filters_converted = sql_filters.convert_filters(
            self._model_table,
            filters,
            session=self._session,
        )
        filters_tuple = (filters_converted,)
        self._where_expression.extend_clauses(filters_tuple)
        return self

    def limit(self, value):
        self._limit_condition = Limit(
            value,
            session=self._session,
        )
        return self

    def for_(self, share=False):
        self._for_expression = For(share)
        return self

    def order_by(self, property_name, sort_type="ASC"):
        column = self._model_table.get_column_by_name(
            property_name,
            wrap_alias=False,
        )

        self._order_by_expressions.append(
            OrderByValue(
                column=column,
                sort_type=sort_type,
                session=self._session,
            )
        )
        return self

    def _build_table_alias_name(self):
        return "t%d" % self._get_inc()

    def _get_inc(self):
        with self._autoinc_lock:
            self._autoinc += 1
            return self._autoinc

    def compile(self):
        # noinspection SqlInjection
        expression = "SELECT %s FROM %s" % (
            ", ".join([exp.compile() for exp in self._select_expressions]),
            " ".join([tbl.compile() for tbl in self._table_references]),
        )
        where_expressions = self._where_expression.construct_expression()
        if where_expressions:
            expression += " WHERE " + where_expressions
        if self._order_by_expressions:
            expression += " ORDER BY %s" % ", ".join(
                [exp.compile() for exp in self._order_by_expressions]
            )
        if self._limit_condition:
            expression += " %s" % self._limit_condition.compile()
        if self._for_expression:
            expression += " %s" % self._for_expression.compile()
        return expression

    def values(self):
        # TODO(efrolov): Must be read only list
        return self._where_expression.value

    def parse_row(self, row):
        return self._result_parser.root.parse(row)

    def parse_results(self, rows):
        return [self.parse_row(row) for row in rows]


class Q(object):
    @staticmethod
    def select(model, session):
        return SelectQ(model, session)



# Copyright 2023 Aleksandr Bochkarev
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import mock

from restalchemy.api.middlewares import retry_on_error
from restalchemy.storage import exceptions as ra_exceptions
from restalchemy.tests.unit import base

from webob import request


class FakeDeadLock(ra_exceptions.DeadLock):
    message = "Deadlock is found"


class FakeRecordNotFound(ra_exceptions.RecordNotFound):
    message = "Can't found record in storage"


class RetryOnErrorMiddlewareTestCase(base.BaseTestCase):
    call_count = 3
    app = "application"

    def get_middlew(self, exc=FakeDeadLock, call_count=None):
        return retry_on_error.RetryOnErrorsMiddleware(
            self.app, exc, call_count or self.call_count
        )

    def test_retry_on_deadlock_exc(self):
        request_mock = mock.Mock(spec=request.Request)
        request_mock.get_response.side_effect = FakeDeadLock
        with self.assertRaises(ra_exceptions.DeadLock):
            self.get_middlew().process_request(request_mock)
        request_mock.get_response.assert_called_with(self.app)
        self.assertEqual(self.call_count, request_mock.get_response.call_count)

    def test_retry_on_record_not_found_exc(self):
        request_mock = mock.Mock(spec=request.Request)
        request_mock.get_response.side_effect = FakeRecordNotFound
        with self.assertRaises(ra_exceptions.RecordNotFound):
            self.get_middlew().process_request(request_mock)
        request_mock.get_response.assert_called_with(self.app)
        self.assertEqual(1, request_mock.get_response.call_count)

    def test_retry_when_no_exc(self):
        request_mock = mock.Mock(spec=request.Request)
        response = self.get_middlew().process_request(request_mock)
        self.assertIsNotNone(response)
        request_mock.get_response.assert_called_with(self.app)
        self.assertEqual(1, request_mock.get_response.call_count)

    def test_retry_when_retry_count_less_0(self):
        request_mock = mock.Mock(spec=request.Request)
        request_mock.get_response.side_effect = FakeDeadLock
        with self.assertRaises(ra_exceptions.DeadLock):
            self.get_middlew(call_count=-1).process_request(request_mock)
        request_mock.get_response.assert_called_with(self.app)
        self.assertEqual(1, request_mock.get_response.call_count)



# Copyright 2019 Eugene Frolov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from restalchemy.storage.sql import engines
from restalchemy.tests.unit import base


class TestEngineTestCase(base.BaseTestCase):
    @mock.patch("mysql.connector.pooling.MySQLConnectionPool")
    def setUp(self, *args):
        super(TestEngineTestCase, self).setUp()
        self._engine = engines.MySQLEngine(db_url="mysql://test:test@test/test")

    def tearDown(self):
        super(TestEngineTestCase, self).tearDown()
        del self._engine

    def test_session_manager_session_as_argument(self):
        session = mock.Mock()

        with self._engine.session_manager(session=session) as s:
            self.assertEqual(s, session)

    def test_session_manager_session_as_thread_storage(self):
        session = mock.Mock()

        with mock.patch.object(
            self._engine, "_get_session_from_storage", return_value=session
        ):
            with self._engine.session_manager() as s:
                self.assertEqual(s, session)

    def test_session_manager_get_new_session(self):
        session = mock.Mock()

        with mock.patch.object(self._engine, "get_session", return_value=session):
            with self._engine.session_manager() as s:
                self.assertEqual(s, session)


class DBConnectionUrlTestCase(base.BaseTestCase):
    """Test case for DBConnectionUrl instance"""

    _DB_URL_TEMPLATE = "mysql://john%s10.0.0.1/mydb"
    _DB_URL_CENSORED = _DB_URL_TEMPLATE % engines.DBConnectionUrl._CENSORED

    def test_repr_with_password(self):
        db_url = engines.DBConnectionUrl(self._DB_URL_TEMPLATE % ":my_cool_secret@")

        actual_repr = repr(db_url)
        actual_str = str(db_url)

        self.assertEqual(actual_repr, self._DB_URL_CENSORED)
        self.assertEqual(actual_str, actual_repr)

    def test_repr_with_empty_password(self):
        db_url = engines.DBConnectionUrl(self._DB_URL_TEMPLATE % ":@")

        actual_repr = repr(db_url)
        actual_str = str(db_url)

        self.assertEqual(actual_repr, self._DB_URL_CENSORED)
        self.assertEqual(actual_str, actual_repr)

    def test_repr_without_password(self):
        db_url = engines.DBConnectionUrl(self._DB_URL_TEMPLATE % "@")

        actual_repr = repr(db_url)
        actual_str = str(db_url)

        self.assertEqual(actual_repr, self._DB_URL_CENSORED)
        self.assertEqual(actual_str, actual_repr)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import copy
import datetime
import logging
import orjson
import uuid

from restalchemy.common import exceptions as ra_exc
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types

LOG = logging.getLogger(__name__)


def build_prop_kwargs(kwargs):
    """
    Build the openapi schema properties from the given ``kwargs``.

    The main purpose of this function is to map the ``kwargs`` to the
    corresponding openapi property name and handle the special cases for
    values that are not serializable to json (like UUIDs).

    :param kwargs: dictionary of keyword arguments
    :return: dictionary of openapi properties
    """
    result = {}
    for k, v in types.KWARGS_OPENAPI_MAP.items():
        if k in kwargs.keys():
            value = kwargs[k]() if callable(kwargs[k]) else kwargs[k]
            if isinstance(value, (types.UUID, uuid.UUID)):
                value = "uuid"
            elif isinstance(value, datetime.datetime):
                value = value.strftime(types.OPENAPI_FMT)
            if hasattr(value, "__dict__"):
                value = {
                    name: prop.get_property_type().to_simple_type(value[name])
                    for name, prop in value.properties.properties.items()
                }
            result[v] = value
    return result


class UnknownType(Exception):
    pass


class AbstractKindType(types.BasePythonType):
    def from_unicode(self, value):
        """
        Deserialize the given value into a python object.

        This method is used by the ``from_unicode`` method of the
        ``BaseType`` class. It takes a unicode string as an argument and
        returns a python object.

        The default implementation of this method simply calls the
        ``from_unicode`` method of the ``BaseType`` class with the
        deserialized value from the given ``value``.

        :param value: unicode string to be deserialized
        :return: python object
        """
        return super().from_unicode(self.from_simple_type(orjson.loads(value)))


class AbstractKindModel(models.Model, metaclass=abc.ABCMeta):
    DOMAIN = uuid.UUID("0188bbae-7e1b-11ee-a337-047c160cda6f")

    @classmethod
    @abc.abstractmethod
    def KIND(cls):
        """
        The kind of the object.

        This abstract property should be overridden by subclasses to return
        a string that uniquely identifies the kind of the object.

        :return: string
        :raises NotImplementedError: If the method is not implemented by a
            subclass.
        """
        raise NotImplementedError()

    kind = properties.property(
        types.String(min_length=1, max_length=128),
        required=True,
        read_only=True,
    )

    def __init__(self, kind=None, **kwargs):
        """
        Initialize the object.

        :param kind: the kind of the object. If not given, the class' KIND
            property is used.
        :param kwargs: keyword arguments to be passed to the parent class'
            constructor.
        """
        kind = kind or self.KIND
        super().__init__(kind=kind, **kwargs)


class KindModelType(types.BasePythonType):
    def __init__(self, model):
        """
        Initialize the type.

        :param model: the model class to be represented by this type.
        :type model: type
        """
        super(KindModelType, self).__init__(
            python_type=model,
            openapi_type="object",
        )

    @property
    def kind(self):
        """
        Return the kind of the object.

        This property is read-only and returns the kind of the object as given
        by the KIND property of the class.

        :return: string
        """
        return self._python_type.KIND

    def from_simple_type(self, value):
        """
        Convert a simple type representation into an instance of the model
        type.

        This function takes a dictionary representation of the model and
        converts it into an instance of the associated model type. Each field
        in the input dictionary is processed using the corresponding
        property's `from_simple_type` method. If there are any extra fields
        in the input dictionary that do not correspond to the model's
        properties, a `ValueError` is raised.

        :param value: A dictionary where keys are field names and values are
            the simple type representations of the model's properties.
        :return: An instance of the model type with fields populated from the
            input dictionary.
        :raises ValueError: If the input dictionary contains unknown fields.
        """

        copied_value = copy.deepcopy(value)
        parsed_value = {}
        for name, prop in self._python_type.properties.properties.items():
            if name in copied_value:
                property_type = prop.get_property_type()
                try:
                    val = copied_value.pop(name)
                    parsed_value[name] = property_type.from_simple_type(val)
                except (ValueError, TypeError):
                    raise ra_exc.ParseError(value="%s=%s" % (name, val))
                except ra_exc.ParseError as e:
                    raise ra_exc.ParseError(value="%s=%s" % (name, e.value))
        if copied_value:
            raise ra_exc.ParseError(value="(Unknown fields: %s)" % (copied_value))
        return self._python_type(**parsed_value)

    def to_simple_type(self, value):
        """
        Convert an instance of the model type to a simple type representation.

        This function takes an instance of the model type and converts it into
        a dictionary representation. Each field in the model is processed
        using the corresponding property's `to_simple_type` method to obtain
        its simple type representation.

        :param value: An instance of the model type with fields to be
            converted.
        :return: A dictionary where keys are field names and values are the
            simple type representations of the model's properties.
        """
        return {
            name: prop.get_property_type().to_simple_type(value[name])
            for name, prop in value.properties.properties.items()
        }

    def to_openapi_spec(self, prop_kwargs):
        """
        Return an OpenAPI specification for this type.

        :param prop_kwargs: additional keyword arguments to be passed to
            `build_prop_kwargs`.
        :return: an OpenAPI specification for this type.
        """
        props = {
            name: prop.get_property_type().to_openapi_spec(prop.get_kwargs())
            for name, prop in self._python_type.properties.properties.items()
        }
        props["kind"] = {"type": "string", "enum": [self.kind]}
        spec = {"type": self.openapi_type, "properties": props}
        spec.update(build_prop_kwargs(kwargs=prop_kwargs))
        return spec


class KindModelSelectorType(types.BaseType):
    def __init__(self, *args):
        """
        Initialize the type.

        :param args: A variable number of arguments
            of :py:class:`KindModelType` instances.
        """
        super(KindModelSelectorType, self).__init__(openapi_type="object")
        self._kind_type_map = {model.kind: model for model in args}

    def validate(self, value):
        """
        Check if the given value is valid according to the type definition.

        This function takes a value and checks if it is valid according to the
        definition of this type. This is done by looking up the type of the
        value in the map of types and checking if the value is valid according
        to that type.

        :param value: The value to be validated.
        :return: ``True`` if the value is valid, ``False`` otherwise.
        """
        try:
            value_type_name = value.kind
            value_type = self._kind_type_map[value_type_name]
            return value_type.validate(value)
        except (AttributeError, KeyError):
            return False

    def to_simple_type(self, value):
        """
        Convert an instance of the model type to a simple type representation.

        This function takes an instance of the model type and converts it into
        a dictionary representation. The dictionary representation is obtained
        by calling the `to_simple_type` method of the type corresponding to
        the kind of the model instance.

        :param value: An instance of the model type with fields to be
            converted.
        :return: A dictionary where keys are field names and values are the
            simple type representations of the model's properties.
        """
        value_type_name = value.kind
        value_type = self._kind_type_map[value_type_name]
        return value_type.to_simple_type(value)

    def from_simple_type(self, value):
        """
        Convert a simple type representation into an instance of the model
        type.

        This function takes a dictionary representation of the model and
        determines the appropriate model class to use based on the 'kind'
        field. It then uses the `from_simple_type` method of the identified
        model class to instantiate an instance of the model type. If the
        'kind' is not found or is invalid, an `UnknownType` error is raised.

        :param value: A dictionary with a 'kind' field indicating the model
            type and additional fields representing the simple type data.
        :return: An instance of the model type corresponding to the 'kind'
            field.
        :raises UnknownType: If the 'kind' field does not correspond to any
            known model type.
        """
        try:
            value_type_name = value["kind"]
            value_class = self._kind_type_map[value_type_name]
            return value_class.from_simple_type(value)
        except (AttributeError, KeyError):
            raise UnknownType("Unknown kind for value: %s" % value)

    def from_unicode(self, value):
        """
        Deserialize the given Unicode string into an instance of the model
        type.

        This method parses the input Unicode string as JSON and converts the
        resulting dictionary into an instance of the corresponding model type
        using the `from_simple_type` method.

        :param value: A Unicode string representing the JSON-encoded simple
            type data.
        :return: An instance of the model type.
        :raises orjson.JSONDecodeError: If the input string is not valid JSON.
        :raises UnknownType: If the 'kind' field in the decoded JSON does not
            correspond to any known model type.
        """

        return self.from_simple_type(orjson.loads(value))

    def to_openapi_spec(self, prop_kwargs):
        """
        Generate an OpenAPI specification for the model type.

        This method creates a dictionary representing the OpenAPI
        specification for the model type. The specification includes the
        OpenAPI type of the model and incorporates additional properties
        from `prop_kwargs` if provided.

        :param prop_kwargs: Additional keyword arguments for building the
            OpenAPI specification.
        :return: A dictionary representing the OpenAPI specification.
        """

        spec = {
            "type": self.openapi_type,
            "oneOf": [
                subtype.to_openapi_spec({}) for subtype in self._kind_type_map.values()
            ],
        }

        spec.update(build_prop_kwargs(kwargs=prop_kwargs))
        return spec



# Copyright 2018 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import uuid

from restalchemy.dm import filters as dm_filters
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import relationships
from restalchemy.dm import types
from restalchemy.storage.sql import filters
from restalchemy.storage.sql import orm
from restalchemy.tests import fixtures
from restalchemy.tests.unit import base
from restalchemy.tests.unit.storage.sql import common


class MultiDict(dict):
    """MultiDict implementation allowing multiple values per key."""

    def __init__(self, *args, **kwargs):
        super().__init__()
        self._data = {}  # Store as {key: [value1, value2, ...]}
        if args or kwargs:
            # Initialize with provided data
            if args:
                other = args[0]
                if hasattr(other, "items"):
                    for key, value in other.items():
                        self.add(key, value)
                else:
                    for key, value in other:
                        self.add(key, value)
            for key, value in kwargs.items():
                self.add(key, value)

    def add(self, key, value):
        """Add value for key, append if key exists."""
        if key in self._data:
            self._data[key].append(value)
        else:
            self._data[key] = [value]

    def items(self):
        """Return all key-value pairs with duplicates."""
        for key, values in self._data.items():
            for value in values:
                yield key, value

    def keys(self):
        """Return all keys with potential duplicates."""
        for key, values in self._data.items():
            for _ in values:
                yield key

    def values(self):
        """Return all values."""
        for values in self._data.values():
            yield from values

    def __getitem__(self, key):
        """Get the list of values for a key."""
        return self._data[key]

    def __setitem__(self, key, value):
        """Set a single value for a key (replaces all existing values)."""
        self._data[key] = [value]

    def __contains__(self, key):
        return key in self._data

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return sum(len(values) for values in self._data.values())

    def get(self, key, default=None):
        """Get the list of values for a key."""
        return self._data.get(key, default)


# Alias for backward compatibility with HTTPHeaderDict
HTTPHeaderDict = MultiDict


TEST_NAME = "FAKE_NAME"
TEST_VALUE = "FAKE_VALUE"
TEST_UUID = uuid.UUID("89d423c5-4365-4be2-bde9-2730909a9af8")


class BaseModel(models.Model):
    name1 = properties.property(types.Integer())
    name2 = properties.property(types.Integer())


class BaseModelWithUUID(BaseModel, models.ModelWithUUID, orm.SQLStorableMixin):
    pass


class BaseModelWithRelation(BaseModel, models.ModelWithUUID):
    parent = relationships.relationship(BaseModelWithUUID)


class EQTestCase(base.BaseTestCase):
    def setUp(self):
        self._expr = filters.EQ(
            column=TEST_NAME,
            value_type=common.AsIsType(),
            value=TEST_VALUE,
            session=fixtures.SessionFixture(),
        )

    def test_construct_expression(self):

        result = self._expr.construct_expression()

        self.assertEqual(TEST_NAME + " = %s", result)

    def test_value_property(self):
        self.assertEqual(self._expr.value, TEST_VALUE)


class NETestCase(base.BaseTestCase):
    def setUp(self):
        self._expr = filters.NE(
            column=TEST_NAME,
            value_type=common.AsIsType(),
            value=TEST_VALUE,
            session=fixtures.SessionFixture(),
        )

    def test_construct_expression(self):

        result = self._expr.construct_expression()

        self.assertEqual(TEST_NAME + " <> %s", result)

    def test_value_property(self):
        self.assertEqual(self._expr.value, TEST_VALUE)


class GTTestCase(base.BaseTestCase):
    def setUp(self):
        self._expr = filters.GT(
            column=TEST_NAME,
            value_type=common.AsIsType(),
            value=TEST_VALUE,
            session=fixtures.SessionFixture(),
        )

    def test_construct_expression(self):

        result = self._expr.construct_expression()

        self.assertEqual(TEST_NAME + " > %s", result)

    def test_value_property(self):
        self.assertEqual(self._expr.value, TEST_VALUE)


class GETestCase(base.BaseTestCase):
    def setUp(self):
        self._expr = filters.GE(
            column=TEST_NAME,
            value_type=common.AsIsType(),
            value=TEST_VALUE,
            session=fixtures.SessionFixture(),
        )

    def test_construct_expression(self):

        result = self._expr.construct_expression()

        self.assertEqual(TEST_NAME + " >= %s", result)

    def test_value_property(self):
        self.assertEqual(self._expr.value, TEST_VALUE)


class LTTestCase(base.BaseTestCase):
    def setUp(self):
        self._expr = filters.LT(
            column=TEST_NAME,
            value_type=common.AsIsType(),
            value=TEST_VALUE,
            session=fixtures.SessionFixture(),
        )

    def test_construct_expression(self):

        result = self._expr.construct_expression()

        self.assertEqual(TEST_NAME + " < %s", result)

    def test_value_property(self):
        self.assertEqual(self._expr.value, TEST_VALUE)


class LETestCase(base.BaseTestCase):
    def setUp(self):
        self._expr = filters.LE(
            column=TEST_NAME,
            value_type=common.AsIsType(),
            value=TEST_VALUE,
            session=fixtures.SessionFixture(),
        )

    def test_construct_expression(self):

        result = self._expr.construct_expression()

        self.assertEqual(TEST_NAME + " <= %s", result)

    def test_value_property(self):
        self.assertEqual(self._expr.value, TEST_VALUE)


class InTestCase(base.BaseTestCase):
    TEST_LIST_VALUES = [1, 2, 3]

    def setUp(self):
        self._expr = filters.MySqlIn(
            column=TEST_NAME,
            value_type=common.AsIsType(),
            value=self.TEST_LIST_VALUES,
            session=fixtures.SessionFixture(),
        )

    def test_construct_expression(self):

        result = self._expr.construct_expression()

        self.assertEqual(TEST_NAME + " IN %s", result)

    def test_value_property(self):
        self.assertEqual(self._expr.value, self.TEST_LIST_VALUES)


class NotInTestCase(base.BaseTestCase):
    TEST_LIST_VALUES = [1, 2, 3]

    def setUp(self):
        self._expr = filters.MySqlNotIn(
            column=TEST_NAME,
            value_type=common.AsIsType(),
            value=self.TEST_LIST_VALUES,
            session=fixtures.SessionFixture(),
        )

    def test_construct_expression(self):

        result = self._expr.construct_expression()

        self.assertEqual(TEST_NAME + " NOT IN %s", result)

    def test_value_property(self):
        self.assertEqual(self._expr.value, self.TEST_LIST_VALUES)


class InEmptyListTestCase(base.BaseTestCase):
    def setUp(self):
        self._expr = filters.MySqlIn(
            column=TEST_NAME,
            value_type=common.AsIsType(),
            value=[],
            session=fixtures.SessionFixture(),
        )

    def test_construct_expression(self):

        result = self._expr.construct_expression()

        self.assertEqual(TEST_NAME + " IN %s", result)

    def test_value_property(self):
        self.assertEqual(self._expr.value, [None])


class IsTestCase(base.BaseTestCase):
    def setUp(self):
        self._expr = filters.Is(
            column=TEST_NAME,
            value_type=common.AsIsType(),
            value=TEST_VALUE,
            session=fixtures.SessionFixture(),
        )

    def test_construct_expression(self):

        result = self._expr.construct_expression()

        self.assertEqual(TEST_NAME + " IS %s", result)

    def test_value_property(self):
        self.assertEqual(self._expr.value, TEST_VALUE)


class IsNotTestCase(base.BaseTestCase):
    def setUp(self):
        self._expr = filters.IsNot(
            column=TEST_NAME,
            value_type=common.AsIsType(),
            value=TEST_VALUE,
            session=fixtures.SessionFixture(),
        )

    def test_construct_expression(self):

        result = self._expr.construct_expression()

        self.assertEqual(TEST_NAME + " IS NOT %s", result)

    def test_value_property(self):
        self.assertEqual(self._expr.value, TEST_VALUE)


class IsPgTestCase(base.BaseTestCase):
    def setUp(self):
        self._expr = filters.PostgreSqlIs(
            column=TEST_NAME,
            value_type=common.AsIsType(),
            value=TEST_VALUE,
            session=fixtures.SessionFixture(),
        )

    def test_construct_expression(self):

        result = self._expr.construct_expression()

        self.assertEqual(TEST_NAME + " IS NOT DISTINCT FROM (%s)", result)

    def test_value_property(self):
        self.assertEqual(self._expr.value, TEST_VALUE)


class IsNotPgTestCase(base.BaseTestCase):
    def setUp(self):
        self._expr = filters.PostgreSqlIsNot(
            column=TEST_NAME,
            value_type=common.AsIsType(),
            value=TEST_VALUE,
            session=fixtures.SessionFixture(),
        )

    def test_construct_expression(self):

        result = self._expr.construct_expression()

        self.assertEqual(TEST_NAME + " IS DISTINCT FROM (%s)", result)

    def test_value_property(self):
        self.assertEqual(self._expr.value, TEST_VALUE)


class LikeTestCase(base.BaseTestCase):
    def setUp(self):
        self._expr = filters.Like(
            column=TEST_NAME,
            value_type=common.AsIsType(),
            value=TEST_VALUE,
            session=fixtures.SessionFixture(),
        )

    def test_construct_expression(self):

        result = self._expr.construct_expression()

        self.assertEqual(TEST_NAME + " LIKE %s", result)

    def test_value_property(self):
        self.assertEqual(self._expr.value, TEST_VALUE)


class NotLikeTestCase(base.BaseTestCase):
    def setUp(self):
        self._expr = filters.NotLike(
            column=TEST_NAME,
            value_type=common.AsIsType(),
            value=TEST_VALUE,
            session=fixtures.SessionFixture(),
        )

    def test_construct_expression(self):

        result = self._expr.construct_expression()

        self.assertEqual(TEST_NAME + " NOT LIKE %s", result)

    def test_value_property(self):
        self.assertEqual(self._expr.value, TEST_VALUE)


class ConvertFiltersTestCase(base.BaseTestCase):
    def test_convert_filters_new(self):
        d = collections.OrderedDict()
        d["name1"] = dm_filters.EQ(1)
        d["name2"] = dm_filters.EQ(2)
        filter_list = dm_filters.AND(d)

        processed = filters.convert_filters(
            BaseModel,
            filter_list,
            session=fixtures.SessionFixture(),
        )

        self.assertEqual(
            "(`name1` = %s AND `name2` = %s)", processed.construct_expression()
        )
        self.assertEqual([1, 2], processed.value)

    def test_convert_filters_new_separate_dicts(self):
        filter_list = dm_filters.AND(
            {"name1": dm_filters.EQ(1)}, {"name2": dm_filters.EQ(2)}
        )

        processed = filters.convert_filters(
            BaseModel,
            filter_list,
            session=fixtures.SessionFixture(),
        )

        self.assertEqual(
            "(`name1` = %s AND `name2` = %s)", processed.construct_expression()
        )
        self.assertEqual([1, 2], processed.value)

    def test_convert_filters_new_nested(self):
        d = collections.OrderedDict()
        d["name1"] = dm_filters.EQ(1)
        d["name2"] = dm_filters.EQ(2)
        filter_list = dm_filters.OR(
            dm_filters.AND(d), dm_filters.AND({"name2": dm_filters.EQ(2)})
        )

        processed = filters.convert_filters(
            BaseModel,
            filter_list,
            session=fixtures.SessionFixture(),
        )

        self.assertEqual(
            "((`name1` = %s AND `name2` = %s) OR `name2` = %s)",
            processed.construct_expression(),
        )
        self.assertEqual([1, 2, 2], processed.value)

    def test_convert_filters_old(self):
        d = collections.OrderedDict()
        d["name1"] = dm_filters.EQ(1)
        d["name2"] = dm_filters.EQ(2)
        filter_list = d

        processed = filters.convert_filters(
            BaseModel,
            filter_list,
            session=fixtures.SessionFixture(),
        )

        self.assertEqual(
            "(`name1` = %s AND `name2` = %s)", processed.construct_expression()
        )
        self.assertEqual([1, 2], processed.value)

    def test_convert_filters_old_multidict(self):
        d = HTTPHeaderDict()
        d.add("name1", dm_filters.EQ(1))
        d.add("name1", dm_filters.EQ(1))
        filter_list = d

        processed = filters.convert_filters(
            BaseModel,
            filter_list,
            session=fixtures.SessionFixture(),
        )

        self.assertEqual(
            "(`name1` = %s AND `name1` = %s)", processed.construct_expression()
        )
        self.assertEqual([1, 1], processed.value)

    def test_convert_filters_new_relationship_by_model(self):
        model = BaseModelWithUUID(name1=1, name2=2, uuid=TEST_UUID)
        filter_list = {"parent": dm_filters.EQ(model)}

        processed = filters.convert_filters(
            BaseModelWithRelation,
            filter_list,
            session=fixtures.SessionFixture(),
        )

        self.assertEqual("`parent` = %s", processed.construct_expression())
        self.assertEqual([str(TEST_UUID)], processed.value)

    def test_convert_filters_new_relationship_by_id(self):
        filter_list = {"parent": dm_filters.EQ(TEST_UUID)}

        processed = filters.convert_filters(
            BaseModelWithRelation,
            filter_list,
            session=fixtures.SessionFixture(),
        )

        self.assertEqual("`parent` = %s", processed.construct_expression())
        self.assertEqual([str(TEST_UUID)], processed.value)



# Copyright 2019 Eugene Frolov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import os

import mock

from restalchemy.dm import filters
from restalchemy.storage.sql import migrations as sql_migrations
from restalchemy.tests.functional import base

INIT_MIGRATION = "0000-init-0d06a9"
FIRST_MIGRATION = "0001-first-fc0c16"
SECOND_MIGRATION = "0002-second-562b5a"
THIRD_MIGRATION = "0003-third-bbd5d8"

HEAD_MIGRATION = THIRD_MIGRATION

NEW_MIGRATION_NUMBER = "0004"
NEW_MIGRATION_MESSAGE = "fourth"
NEW_MIGRATION_DEPENDS = [HEAD_MIGRATION]

MIGRATIONS_TOTAL_COUNT = len(
    [INIT_MIGRATION, FIRST_MIGRATION, SECOND_MIGRATION, THIRD_MIGRATION]
)

MIGRATIONS_FIXTURES_DIR_NAME = "migrations_fixtures"
NONEXISTENT_MIGRATION = "nonexistent_migration"


def ensure_py_extension(filename):
    py_extension = ".py"
    if filename.endswith(py_extension):
        return filename
    return filename + py_extension


def get_filename_hash(filename):
    _, file = os.path.split(filename)
    return file.split("-")[-1].split(".")[0]


def test_get_filename_hash():
    testpath = "/very/long/path/with/file/0000-message-0d6a9.py"
    shortpath = "./0000-message-0d6a9.py"
    onlyname = "0000-message-0d6a9.py"
    longfilename = "0000-very-very-long-message-0d6a9.py"
    wrongname = "0d6a9-0000-message.py"

    expected_hash = "0d6a9"

    assert get_filename_hash(testpath) == expected_hash
    assert get_filename_hash(shortpath) == expected_hash
    assert get_filename_hash(onlyname) == expected_hash
    assert get_filename_hash(longfilename) == expected_hash
    assert get_filename_hash(wrongname) != expected_hash


class BaseMigrationTestCase(base.BaseDBEngineTestCase):
    def setUp(self):
        super(BaseMigrationTestCase, self).setUp()

        self.migration_engine = self.get_migration_engine()

        self._drop_ra_migrations_table()

    def tearDown(self):
        super(BaseMigrationTestCase, self).tearDown()

        self._drop_ra_migrations_table()

    @staticmethod
    def get_migration_engine(migrations_dir_name="migrations"):
        migrations_path = os.path.join(
            os.path.dirname(__file__),
            MIGRATIONS_FIXTURES_DIR_NAME,
            migrations_dir_name,
        )
        migration_engine = sql_migrations.MigrationEngine(
            migrations_path=migrations_path
        )
        return migration_engine

    def _truncate_ra_migrations_table(self):
        with self.engine.session_manager() as session:
            self.truncate_table(sql_migrations.RA_MIGRATION_TABLE_NAME, session=session)

    def _drop_ra_migrations_table(self):
        with self.engine.session_manager() as session:
            self.drop_table(sql_migrations.RA_MIGRATION_TABLE_NAME, session=session)

    def load_migrations(self):
        with self.engine.session_manager() as session:
            migrations = self.migration_engine._load_migration_controllers(session)
        return migrations

    def init_migration_table(self):
        with self.engine.session_manager() as session:
            self.migration_engine._init_migration_table(session)


class MigrationsModelTestCase(BaseMigrationTestCase):
    def test_instantiate_migration_model(self):
        model_cls = sql_migrations.MigrationModel

        self.assertIsInstance(model_cls(), model_cls)

    def test_migration_already_applied(self):

        self.migration_engine.apply_migration(migration_name=FIRST_MIGRATION)

        with mock.patch.object(logging.Logger, "warning") as warning:
            self.migration_engine.apply_migration(migration_name=FIRST_MIGRATION)
            warning.assert_called_with(
                "Migration '%s' is already applied", FIRST_MIGRATION
            )

    def test_migration_not_applied(self):

        with mock.patch.object(logging.Logger, "warning") as warning:
            self.migration_engine.rollback_migration(migration_name=FIRST_MIGRATION)
            warning.assert_called_with("Migration '%s' is not applied", FIRST_MIGRATION)

    def test_migration_in_db_is_correct(self):

        self.migration_engine.apply_migration(migration_name=FIRST_MIGRATION)

        db_migrations = sql_migrations.MigrationModel.objects.get_all()
        self.assertTrue(all([m.applied for m in db_migrations]))
        self.assertEqual(len(db_migrations), 2)

        self.migration_engine.rollback_migration(migration_name=FIRST_MIGRATION)

        # Only one applied init migration after rollback
        db_filter = {"applied": filters.EQ(True)}
        m = sql_migrations.MigrationModel.objects.get_one(filters=db_filter)
        hash_len = self.migration_engine.FILENAME_HASH_LEN
        self.assertEqual(str(m.uuid)[:hash_len], INIT_MIGRATION[-hash_len:])

    def test_migration_head_is_latest(self):
        expected_uuids = [
            "0d06a988-90cc-48ab-a842-b979cdf8975d",
            "fc0c165e-9c69-4e47-b7e3-0bc3a2bebfab",
            "562b5a12-cb70-4f77-896b-3a6cab7c3019",
            "bbd5d871-4b0e-4856-b56e-95b2abb7cf48",
        ]
        with self.engine.session_manager() as session:
            self.migration_engine._init_migration_table(session)

        db_filter = {"applied": filters.EQ(True)}
        db_migrations = sql_migrations.MigrationModel.objects.get_all(filters=db_filter)
        self.assertEqual(0, len(db_migrations))

        latest_migration_name = self.migration_engine.get_latest_migration()

        self.migration_engine.apply_migration(migration_name=latest_migration_name)
        db_migrations = sql_migrations.MigrationModel.objects.get_all(filters=db_filter)
        self.assertEqual(4, len(db_migrations))
        self.assertTrue(
            str(migration.uuid) in expected_uuids for migration in db_migrations
        )

    def test_find_head_in_two_migration_sequences(self):
        # test valid migrations
        #
        # migrations dependencies:
        # 0000-init.py <- 0001-first.py
        # 0002-second.py(MANUAL) <- 0003-third.py(MANUAL)
        # Expected last migration: 0001-first.py
        expected_last_migration = "0001-first-a8a827.py"
        custom_migration_engine = self.get_migration_engine("migration_ok_1")

        last_migration = custom_migration_engine.get_latest_migration()

        self.assertEqual(expected_last_migration, last_migration)

    def test_find_head_in_two_separate_migrations(self):
        # test valid migrations
        #
        # migrations dependencies:
        # 0000-init.py  0001-first.py(MANUAL)
        # Expected last migration: 0000-init.py
        expected_last_migration = "0000-init-672a1b.py"
        custom_migration_engine = self.get_migration_engine("migration_ok_3")

        last_migration = custom_migration_engine.get_latest_migration()

        self.assertEqual(expected_last_migration, last_migration)

    def test_find_head_in_long_sequence_migrations(self):
        # test valid migrations
        #
        # migrations dependencies:
        #               0000-init.py  0001-first.py(MANUAL)
        # 0004-fourth.py  ^-- 0002-second.py   ^-- 0003-third.py(MANUAL)
        #    ^-- 0005-fifth.py --^        ^-- 0006-sixth.py
        #                ^-- 0007-seventh.py --^
        # Expected last migration: 0007-seventh.py
        expected_last_migration = "0007-seventh-7368be.py"
        custom_migration_engine = self.get_migration_engine("migration_ok_2")

        last_migration = custom_migration_engine.get_latest_migration()

        self.assertEqual(expected_last_migration, last_migration)

    def test_migrations_with_two_head(self):
        # test invalid migrations
        #
        # migrations dependencies:
        # 0000-init.py <- 0001-first.py   0002-second.py
        # Expected: has two last migrations: 0001-first.py, 0002-second.py
        custom_migration_engine = self.get_migration_engine(
            "migrations_invalid_two_last_migrations"
        )

        with self.assertRaises(sql_migrations.HeadMigrationNotFoundException):
            custom_migration_engine.get_latest_migration()

    def test_migrations_depends_from_manual(self):
        # test valid migrations
        #
        # migrations dependencies:
        # 0000-init.py <- 0001-first.py
        #                      ^--  0002-second.py --> 0003-third.py(MANUAL)
        # Expected: 0002-second.py
        expected_last_migration = "0002-second-c9221f.py"
        custom_migration_engine = self.get_migration_engine("migration_ok_4")

        last_migration = custom_migration_engine.get_latest_migration()

        self.assertEqual(expected_last_migration, last_migration)

    def test_not_manual_migration_depends_from_manual(self):
        custom_migration_engine = self.get_migration_engine("migration_ok_1")
        new_migration_depends = ["0003-third-11f1da.py"]
        fmdm = custom_migration_engine.validate_auto_migration_dependencies

        with mock.patch.object(logging.Logger, "warning") as warning:
            result = fmdm(new_migration_depends)
            warning.assert_called_with("Manual migration(s) is(are) in dependencies!")

        self.assertFalse(result)

    def test_not_manual_migration_depends_from_not_manual(self):
        custom_migration_engine = self.get_migration_engine("migration_ok_1")
        new_migration_depends = ["0001-first-a8a827.py"]
        fmdm = custom_migration_engine.validate_auto_migration_dependencies

        with mock.patch.object(logging.Logger, "warning") as warning:
            result = fmdm(new_migration_depends)
            warning.assert_not_called()

        self.assertTrue(result)

    def test_get_unapplied_migrations(self):
        custom_migration_engine = self.get_migration_engine("migration_ok_1")
        expected_result = ["0000-init-1711de.py", "0001-first-a8a827.py"]

        with self.engine.session_manager() as session:
            result = custom_migration_engine.get_unapplied_migrations(
                session=session, include_manual=False
            )
        result = list(result.keys())
        result.sort(key=lambda x: x.split("-")[0])

        self.assertListEqual(expected_result, result)

    def test_get_unapplied_mixed_migrations(self):
        custom_migration_engine = self.get_migration_engine("migration_ok_1")
        expected_result = [
            "0001-first-a8a827.py",
            "0002-second-377e90.py",
            "0003-third-11f1da.py",
        ]
        migration_to_apply = "0000-init-1711de.py"
        custom_migration_engine.apply_migration(migration_name=migration_to_apply)

        with self.engine.session_manager() as session:
            result = custom_migration_engine.get_unapplied_migrations(
                session=session, include_manual=True
            )
        result = list(result.keys())
        result.sort(key=lambda x: x.split("-")[1])

        self.assertListEqual(expected_result, result)

    def test_no_unapplied_migrations(self):
        custom_migration_engine = self.get_migration_engine("migration_ok_1")
        head_migration = "0001-first-a8a827.py"
        custom_migration_engine.apply_migration(migration_name=head_migration)
        expected_result = []

        with self.engine.session_manager() as session:
            result = custom_migration_engine.get_unapplied_migrations(
                session=session, include_manual=False
            )

        self.assertListEqual(expected_result, list(result.keys()))

    def test_get_unapplied_manual_migrations(self):
        custom_migration_engine = self.get_migration_engine("migration_ok_1")
        head_migration = "0001-first-a8a827.py"
        custom_migration_engine.apply_migration(migration_name=head_migration)
        expected_result = ["0002-second-377e90.py", "0003-third-11f1da.py"]

        with self.engine.session_manager() as session:
            result = custom_migration_engine.get_unapplied_migrations(
                session=session, include_manual=True
            )
        result = list(result.keys())
        result.sort(key=lambda x: x.split("-")[1])

        self.assertListEqual(expected_result, result)


class MigrationEngineTestCase(BaseMigrationTestCase):
    def test_get_file_name(self):
        file_name = self.migration_engine.get_file_name(FIRST_MIGRATION)

        self.assertEqual("%s.py" % FIRST_MIGRATION, file_name)

    def test_get_file_name_ambiguous_name(self):
        ambiguous_name = "0"

        self.assertRaises(
            ValueError, self.migration_engine.get_file_name, ambiguous_name
        )

    def test_get_file_name_nonexistent(self):

        self.assertRaises(
            ValueError,
            self.migration_engine.get_file_name,
            NONEXISTENT_MIGRATION,
        )

    def test__calculate_depends_head(self):
        depends = [sql_migrations.HEAD_MIGRATION]
        files = self.migration_engine._calculate_depends(depends)
        expected_files = list(
            map(
                ensure_py_extension,
                [
                    HEAD_MIGRATION,
                ],
            )
        )

        self.assertListEqual(expected_files, files)

    def test__calculate_depends_multiple(self):
        depends = [FIRST_MIGRATION, SECOND_MIGRATION]
        files = self.migration_engine._calculate_depends(depends)
        expected_files = list(
            map(
                ensure_py_extension,
                [
                    FIRST_MIGRATION,
                    SECOND_MIGRATION,
                ],
            )
        )

        self.assertListEqual(expected_files, files)

    def test_apply_migration(self):
        with self.engine.session_manager() as session:
            self.migration_engine._init_migration_table(session)
        migrations_before = self.load_migrations()
        self.migration_engine.apply_migration(HEAD_MIGRATION, dry_run=False)
        migrations_after = self.load_migrations()

        self.assertEqual(MIGRATIONS_TOTAL_COUNT, len(migrations_before.keys()))
        self.assertEqual(MIGRATIONS_TOTAL_COUNT, len(migrations_after.keys()))

        # total number of migrations before and after apply should be same
        self.assertEqual(migrations_before.keys(), migrations_after.keys())

        self.assertTrue(
            all([m.is_applied() is False for m in migrations_before.values()])
        )

        self.assertTrue(
            all([m.is_applied() is True for m in migrations_after.values()])
        )

    def test_apply_migration_dry_run(self):
        with self.engine.session_manager() as session:
            self.migration_engine._init_migration_table(session)
        migrations_before = self.load_migrations()

        self.migration_engine.apply_migration(HEAD_MIGRATION, dry_run=True)
        migrations_after = self.load_migrations()

        self.assertEqual(MIGRATIONS_TOTAL_COUNT, len(migrations_before.keys()))
        self.assertEqual(MIGRATIONS_TOTAL_COUNT, len(migrations_after.keys()))

        # total number of migrations before and after apply should be same
        self.assertEqual(migrations_before.keys(), migrations_after.keys())

        self.assertTrue(
            all([m.is_applied() is False for m in migrations_before.values()])
        )

        self.assertTrue(
            all([m.is_applied() is False for m in migrations_after.values()])
        )

    def test_rollback_migration(self):
        self.migration_engine.apply_migration(HEAD_MIGRATION, dry_run=False)
        migrations_before = self.load_migrations()

        self.migration_engine.rollback_migration(INIT_MIGRATION, dry_run=False)
        migrations_after = self.load_migrations()

        self.assertEqual(MIGRATIONS_TOTAL_COUNT, len(migrations_before.keys()))
        self.assertEqual(MIGRATIONS_TOTAL_COUNT, len(migrations_after.keys()))

        # total number of migrations before and after rollback should be same
        self.assertEqual(migrations_before.keys(), migrations_after.keys())

        self.assertTrue(
            all([m.is_applied() is True for m in migrations_before.values()])
        )

        self.assertTrue(
            all([m.is_applied() is False for m in migrations_after.values()])
        )

    def test_rollback_migration_dry_run(self):
        self.migration_engine.apply_migration(HEAD_MIGRATION, dry_run=False)
        migrations_before = self.load_migrations()

        self.migration_engine.rollback_migration(INIT_MIGRATION, dry_run=True)
        migrations_after = self.load_migrations()

        self.assertEqual(MIGRATIONS_TOTAL_COUNT, len(migrations_before.keys()))
        self.assertEqual(MIGRATIONS_TOTAL_COUNT, len(migrations_after.keys()))

        # total number of migrations before and after rollback should be same
        self.assertEqual(migrations_before.keys(), migrations_after.keys())

        self.assertTrue(
            all([m.is_applied() is True for m in migrations_before.values()])
        )

        self.assertTrue(
            all([m.is_applied() is True for m in migrations_after.values()])
        )

    @mock.patch("builtins.open", new_callable=mock.mock_open())
    def test_create_new_migration(self, file_mock):

        self.migration_engine.new_migration(
            NEW_MIGRATION_DEPENDS, NEW_MIGRATION_MESSAGE, dry_run=False
        )

        self.assertTrue(file_mock.called)

        # two calls - load template, write new migration
        self.assertEqual(2, file_mock.call_count)

        template_path = os.path.join(
            os.path.dirname(sql_migrations.__file__), "migration_templ.tmpl"
        )

        template_read_args = file_mock.call_args_list[1][0]
        migration_write_args = file_mock.call_args_list[0][0]

        self.assertEqual((template_path, "r"), template_read_args)

        self.assertEqual("w", migration_write_args[1])

        self.assertTrue(
            migration_write_args[0].endswith(
                "%s-%s-%s.py"
                % (
                    NEW_MIGRATION_NUMBER,
                    NEW_MIGRATION_MESSAGE,
                    get_filename_hash(migration_write_args[0]),
                )
            )
        )

    @mock.patch("builtins.open", new_callable=mock.mock_open())
    def test_create_new_migration_manual(self, file_mock):

        self.migration_engine.new_migration(
            NEW_MIGRATION_DEPENDS,
            NEW_MIGRATION_MESSAGE,
            dry_run=False,
            is_manual=True,
        )

        self.assertTrue(file_mock.called)

        # two calls - load template, write new migration
        self.assertEqual(2, file_mock.call_count)

        template_path = os.path.join(
            os.path.dirname(sql_migrations.__file__), "migration_templ.tmpl"
        )

        template_read_args = file_mock.call_args_list[1][0]
        migration_write_args = file_mock.call_args_list[0][0]

        self.assertEqual((template_path, "r"), template_read_args)

        self.assertEqual("w", migration_write_args[1])

        self.assertTrue(
            migration_write_args[0].endswith(
                "%s-%s-%s.py"
                % (
                    sql_migrations.MANUAL_MIGRATION,
                    NEW_MIGRATION_MESSAGE,
                    get_filename_hash(migration_write_args[0]),
                )
            )
        )

    @mock.patch("builtins.open", new_callable=mock.mock_open())
    def test_create_new_migration_dry_run(self, file_mock):
        self.migration_engine.new_migration(
            NEW_MIGRATION_DEPENDS, NEW_MIGRATION_MESSAGE, dry_run=True
        )

        self.assertFalse(file_mock.called)



#    Copyright 2021 George Melikov.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc


class AbstractClause(metaclass=abc.ABCMeta):
    def __init__(self, session):
        self._session = session

    @abc.abstractmethod
    def compile(self):
        raise NotImplementedError()

    @property
    def original(self):
        return self


class BaseAlias(AbstractClause):
    def __init__(self, clause, name, session):
        super(BaseAlias, self).__init__(session)
        self._clause = clause
        self._name = name

    @property
    def name(self):
        return self._name

    @property
    def original_name(self):
        return self.original.name

    @property
    def original(self):
        return self._clause

    def compile(self):
        return "%s AS %s" % (
            self._clause.compile(),
            self._session.engine.escape(self.name),
        )


class ColumnAlias(BaseAlias):
    @property
    def model_property(self):
        return self.original.model_property


class TableAlias(BaseAlias):
    def _wrap(self, column):
        return ColumnAlias(
            column,
            "%s_%s" % (self.name, column.name),
            self._session,
        )

    def get_columns(self, with_prefetch=True, wrap_alias=True):
        return [
            self.get_column_by_name(col.name, wrap_alias)
            for col in self._clause.get_columns(with_prefetch)
        ]

    def get_prefetch_columns(self, wrap_alias=True):
        return [
            self.get_column_by_name(col.name, wrap_alias)
            for col in self._clause.get_prefetch_columns()
        ]

    def get_column_by_name(self, name, wrap_alias=True):
        result = ColumnFullPath(
            self,
            self._clause.get_column_by_name(name),
            session=self._session,
        )
        return self._wrap(result) if wrap_alias else result


class Column(AbstractClause):
    def __init__(self, name, prop, session):
        self._name = name
        self._prop = prop
        super(Column, self).__init__(session=session)

    @property
    def name(self):
        return self._name

    @property
    def model_property(self):
        return self._prop

    @property
    def original_name(self):
        return self.name

    def compile(self):
        return "%s" % (self._session.engine.escape(self._name))


class ColumnFullPath(AbstractClause):
    def __init__(self, table, column, session):
        super(ColumnFullPath, self).__init__(session)
        self._table = table
        self._column = column

    @property
    def name(self):
        return self._column.name

    @property
    def model_property(self):
        return self._column.model_property

    @property
    def original_name(self):
        return self.name

    def compile(self):
        return "%s.%s" % (
            self._session.engine.escape(self._table.name),
            self._column.compile(),
        )



# Copyright 2021 George Melikov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
import re

from restalchemy.dm import types

DNS_LABEL_MAX_LEN = 63
FQDN_MAX_LEN = 254
HOSTNAME_MAX_LEN = FQDN_MAX_LEN - 1
FQDN_MIN_LEVELS = 1
FQDN_TEMPLATE = r"(?=^.{2,%i}$)(^((?!-)[a-zA-Z0-9-_]{1,%i}(?<!-)\.){%i,}$)"
HOSTNAME_MIN_LEVELS = 1
LEADING_UNDERSCORE_GROUP = "(_?)"
UNDERSCORE = "_"
HOSTNAME_TEMPLATE = (
    r"(?=^.{1,%i}$)(^%s((?![-%s])[a-zA-Z0-9-%s]{1,%i}"
    r"(?<![-%s])\.){%i,}((?!-)[a-zA-Z0-9-]{1,%i}(?<!-))$)"
)


class IPAddress(types.BaseType):
    def __init__(self, **kwargs):
        super(IPAddress, self).__init__(openapi_type="string", **kwargs)

    def validate(self, value):
        return isinstance(value, netaddr.IPAddress)

    def to_simple_type(self, value):
        return str(value)

    def from_simple_type(self, value):
        return netaddr.IPAddress(value)

    def from_unicode(self, value):
        return self.from_simple_type(value)

    def to_openapi_spec(self, prop_kwargs):
        spec = {
            "type": self.openapi_type,
            "anyOf": [{"format": "ipv4"}, {"format": "ipv6"}],
        }
        spec.update(
            types.build_prop_kwargs(
                kwargs=prop_kwargs, to_simple_type=self.to_simple_type
            )
        )
        return spec


class Network(types.BaseType):
    def __init__(self, **kwargs):
        super(Network, self).__init__(openapi_type="string", **kwargs)

    def validate(self, value):
        return isinstance(value, netaddr.IPNetwork)

    def to_simple_type(self, value):
        return str(value)

    def from_simple_type(self, value):
        return netaddr.IPNetwork(value).cidr

    def from_unicode(self, value):
        return self.from_simple_type(value)


class IpWithMask(types.BaseType):
    def validate(self, value):
        return isinstance(value, netaddr.IPNetwork)

    def to_simple_type(self, value):
        return str(value)

    def from_simple_type(self, value):
        return netaddr.IPNetwork(value)

    def from_unicode(self, value):
        return self.from_simple_type(value)


class OUI(types.BaseCompiledRegExpTypeFromAttr):
    pattern = re.compile(r"^([0-9a-fA-F]{2,2}:){2,2}[0-9a-fA-F]{2,2}$")


class RecordName(types.BaseCompiledRegExpTypeFromAttr):
    pattern = re.compile(r"^([a-zA-Z0-9-_]{1,61}\.{0,1}){0,30}$")

    def from_simple_type(self, value):
        converted_value = super(RecordName, self).from_simple_type(value)
        return converted_value.rstrip(".").rstrip("@")

    def to_simple_type(self, value):
        converted_value = super(RecordName, self).to_simple_type(value)
        return converted_value if len(converted_value) > 0 else "@"


class RecordNameWithWildcard(RecordName):
    # Difference - allow wildcard at the beginning of domain name.
    pattern = re.compile(r"^(\*\.){0,1}([a-zA-Z0-9-_]{1,61}\.{0,1}){0,30}$")


class SrvName(RecordName):
    def validate(self, value):
        parts = value.split(".")
        if len(parts) < 2:
            return False

        if not parts[0].startswith("_"):
            return False

        if not parts[1].startswith("_"):
            return False

        record_name = ".".join(parts[2:])

        if record_name and not super(SrvName, self).validate(record_name):
            return False

        return True


class FQDN(types.BaseCompiledRegExpTypeFromAttr):
    """FQDN type. Allows 1 level too. Root only is prohibited.

    See https://github.com/powerdns/pdns/blob/master/pdns/dnsname.cc#L44
    and https://github.com/powerdns/pdns/blob/master/pdns/ws-api.cc#L387
    """

    pattern = re.compile(
        FQDN_TEMPLATE % (FQDN_MAX_LEN, DNS_LABEL_MAX_LEN, FQDN_MIN_LEVELS)
    )

    def __init__(self, min_levels=FQDN_MIN_LEVELS, **kwargs):
        if min_levels > FQDN_MIN_LEVELS:
            self.pattern = re.compile(
                FQDN_TEMPLATE % (FQDN_MAX_LEN, DNS_LABEL_MAX_LEN, min_levels)
            )
        super(FQDN, self).__init__(**kwargs)


class Hostname(types.BaseCompiledRegExpTypeFromAttr):
    """Same as FQDN but without root dot. Allows 1 level too."""

    def __init__(
        self,
        min_levels=HOSTNAME_MIN_LEVELS,
        allow_leading_underscore=False,
        allow_middle_underscore=False,
        **kwargs,
    ):
        underscore_or_empty = UNDERSCORE if allow_middle_underscore else ""
        self.pattern = re.compile(
            HOSTNAME_TEMPLATE
            % (
                HOSTNAME_MAX_LEN,
                LEADING_UNDERSCORE_GROUP if allow_leading_underscore else "",
                underscore_or_empty,
                underscore_or_empty,
                DNS_LABEL_MAX_LEN,
                underscore_or_empty,
                max([min_levels, HOSTNAME_MIN_LEVELS]) - 1,
                DNS_LABEL_MAX_LEN,
            )
        )
        super(Hostname, self).__init__(**kwargs)


class IPRange(types.BaseType):
    """IPRange type.

    This type represents an IP range as a string. The string is expected
    to be in the format: <start_ip>-<end_ip>
    where <start_ip> and <end_ip> are valid IPv4 address.
    """

    SEPARATOR = "-"

    def __init__(self, **kwargs):
        super(IPRange, self).__init__(openapi_type="string", **kwargs)

    def validate(self, value):
        return isinstance(value, netaddr.IPRange)

    def to_simple_type(self, value):
        return str(value)

    def from_simple_type(self, value):
        return netaddr.IPRange(*value.split(self.SEPARATOR))

    def from_unicode(self, value):
        return self.from_simple_type(value)



# Copyright 2018 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
from collections import abc as collections_abc
import logging

from restalchemy.dm import filters
from restalchemy.dm import types
from restalchemy.storage.sql.dialect.query_builder import common

LOG = logging.getLogger(__name__)


class AbstractClause(metaclass=abc.ABCMeta):
    def __init__(self, column, value_type, value, session):
        super(AbstractClause, self).__init__()
        self._value = self._convert_value(value_type, value)
        self._column = column
        self._session = session

    def _convert_value(self, value_type, value):
        return value_type.to_simple_type(value)

    @property
    def value(self):
        return self._value

    @property
    def column(self):
        return (
            self._column.compile()
            if isinstance(self._column, common.ColumnFullPath)
            else self._column
        )

    @abc.abstractmethod
    def construct_expression(self):
        raise NotImplementedError()


class EQ(AbstractClause):
    def construct_expression(self):
        return f"{self.column} = %s"


class NE(AbstractClause):
    def construct_expression(self):
        return f"{self.column} <> %s"


class GT(AbstractClause):
    def construct_expression(self):
        return f"{self.column} > %s"


class GE(AbstractClause):
    def construct_expression(self):
        return f"{self.column} >= %s"


class LT(AbstractClause):
    def construct_expression(self):
        return f"{self.column} < %s"


class LE(AbstractClause):
    def construct_expression(self):
        return f"{self.column} <= %s"


class Is(AbstractClause):
    def construct_expression(self):
        return f"{self.column} IS %s"


class IsNot(AbstractClause):
    def construct_expression(self):
        return f"{self.column} IS NOT %s"


class In(AbstractClause):
    def _convert_value(self, value_type, value):
        # Note(efrolov): Replace empty list with [Null]. Some SQL servers
        #                forbid empty lists in "in" operator.
        return [value_type.to_simple_type(item) for item in value] or [None]


class MySqlIn(In):
    def construct_expression(self):
        return f"{self.column} IN %s"


class PostgreSqlIn(In):
    def construct_expression(self):
        return f"{self.column} = ANY(%s)"


class MySqlNotIn(In):
    def construct_expression(self):
        return f"{self.column} NOT IN %s"


class PostgreSqlNotIn(In):
    def construct_expression(self):
        return f"{self.column} != ANY(%s)"


class PostgreSqlIs(Is):
    def construct_expression(self):
        return f"{self.column} IS NOT DISTINCT FROM (%s)"


class PostgreSqlIsNot(IsNot):
    def construct_expression(self):
        return f"{self.column} IS DISTINCT FROM (%s)"


class Like(AbstractClause):
    def construct_expression(self):
        return ("%s LIKE " % self.column) + "%s"


class NotLike(AbstractClause):
    def construct_expression(self):
        return ("%s NOT LIKE " % self.column) + "%s"


class AbstractExpression(metaclass=abc.ABCMeta):
    def __init__(self, *clauses):
        super(AbstractExpression, self).__init__()
        self._clauses = clauses

    def extend_clauses(self, clauses):
        self._clauses = self._clauses + clauses

    @property
    def clauses(self):
        return self._clauses

    @property
    def value(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def construct_expression(self):
        raise NotImplementedError()


class ClauseList(AbstractExpression):
    @property
    def value(self):
        res = []
        for val in self._clauses:
            if isinstance(val, AbstractExpression):
                res.extend(val.value)
            else:
                res.append(val.value)
        return res

    @property
    def operator(self):
        raise NotImplementedError

    def construct_expression(self):
        clause_count = len(self._clauses)

        if clause_count == 0:
            return ""

        if clause_count == 1:
            return self._clauses[0].construct_expression()

        joined = f" {self.operator} ".join(
            clause.construct_expression() for clause in self._clauses
        )
        return f"({joined})"


class AND(ClauseList):
    operator = "AND"


class OR(ClauseList):
    operator = "OR"


FILTER_MAPPING = {
    "mysql": {
        filters.EQ: EQ,
        filters.NE: NE,
        filters.GT: GT,
        filters.GE: GE,
        filters.LE: LE,
        filters.LT: LT,
        filters.Is: Is,
        filters.IsNot: IsNot,
        filters.In: MySqlIn,
        filters.NotIn: MySqlNotIn,
        filters.Like: Like,
        filters.NotLike: NotLike,
    },
    "postgresql": {
        filters.EQ: EQ,
        filters.NE: NE,
        filters.GT: GT,
        filters.GE: GE,
        filters.LE: LE,
        filters.LT: LT,
        filters.Is: PostgreSqlIs,
        filters.IsNot: PostgreSqlIsNot,
        filters.In: PostgreSqlIn,
        filters.NotIn: PostgreSqlNotIn,
        filters.Like: Like,
        filters.NotLike: NotLike,
    },
}

FILTER_EXPR_MAPPING = {
    filters.AND: AND,
    filters.OR: OR,
}


class AsIsType(types.BaseType):
    def validate(self, value):
        return True

    def to_simple_type(self, value):
        return value

    def from_simple_type(self, value):
        return value

    def from_unicode(self, value):
        return value


def convert_filters(model, filters_root, session):
    filters_root = filters_root or filters.AND()
    if isinstance(filters_root, filters.AbstractExpression):
        return iterate_filters(model, filters_root, session=session)
    return iterate_filters(
        model,
        filters.AND(filters_root),
        session=session,
    )


def iterate_filters(model, filter_list, session):
    # Just expression
    if isinstance(filter_list, filters.AbstractExpression):
        clauses = iterate_filters(model, filter_list.clauses, session)
        return FILTER_EXPR_MAPPING[type(filter_list)](*clauses)

    # Tuple of causes from expression
    if isinstance(filter_list, tuple):
        clauses = []
        for cause in filter_list:
            c_causes = iterate_filters(model, cause, session)
            if isinstance(cause, filters.AbstractExpression):
                clauses.append(c_causes)
            else:
                clauses.extend(c_causes)
        return clauses

    # old style mappings (dict, multidict)
    if isinstance(filter_list, collections_abc.Mapping):
        clauses = []
        for name, filt in filter_list.items():
            if isinstance(model, common.TableAlias):
                value_type = (
                    model.original.model.properties.properties[name].get_property_type()
                ) or AsIsType()
                column = model.get_column_by_name(
                    name,
                    wrap_alias=False,
                )
            else:
                value_type = (
                    model.properties.properties[name].get_property_type()
                ) or AsIsType()
                column = session.engine.escape(name)
            # Make API compatible with previous versions.
            if not isinstance(filt, filters.AbstractClause):
                LOG.warning(
                    "DEPRECATED: pleases use %s wrapper for filter value",
                    filters.EQ,
                )
                clauses.append(EQ(column, value_type, filt, session=session))
                continue

            try:
                clauses.append(
                    FILTER_MAPPING[session.engine.dialect.name][type(filt)](
                        column,
                        value_type,
                        filt.value,
                        session=session,
                    ),
                )
            except KeyError:
                raise ValueError(
                    "Can't convert API filter to SQL storage "
                    "filter. Unknown filter %s" % filt
                )
        return clauses

    raise ValueError("Unknown type of filters: %s" % filter_list)



# Copyright 2019 Eugene Frolov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock


class DialectFixture(mock.Mock):
    @property
    def name(self):
        return "mysql"


class EngineFixture(mock.Mock):
    def escape(self, value):
        return f"`{value}`"

    @property
    def dialect(self):
        return DialectFixture()


class SessionFixture(mock.Mock):
    @property
    def engine(self):
        return EngineFixture()

    @engine.setter
    def engine(self, value):
        pass



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import socket
import uuid as pyuuid


import collections
import mock
import requests
from urllib import parse
from webob import request
from parameterized import parameterized


from functools import partial
from restalchemy.api import constants
from restalchemy.api import packers
from restalchemy.api import resources
from restalchemy.common import utils
from restalchemy.dm import filters
from restalchemy.dm import types
from restalchemy.storage import exceptions
from restalchemy.storage.sql.dialect import exceptions as dialect_exc
from restalchemy.storage.sql.tables import SQLTable
from restalchemy.tests.functional import base
from restalchemy.tests.functional.restapi.ra_based.microservice import (
    storable_models as models,
)
from restalchemy.tests.functional.restapi.ra_based.microservice import routes
from restalchemy.tests.functional.restapi.ra_based.microservice import service
from restalchemy.tests.utils import make_test_name

TEMPL_SERVICE_ENDPOINT = utils.lastslash("http://127.0.0.1:%s/")
TEMPL_ROOT_COLLECTION_ENDPOINT = TEMPL_SERVICE_ENDPOINT
TEMPL_SPEC_SPEC_ENDPOINT = utils.lastslash(
    parse.urljoin(TEMPL_SERVICE_ENDPOINT, "specifications")
)
TEMPL_OPENAPI_SPEC_ENDPOINT = parse.urljoin(TEMPL_SPEC_SPEC_ENDPOINT, "3.0.3")
TEMPL_V1_COLLECTION_ENDPOINT = utils.lastslash(
    parse.urljoin(TEMPL_SERVICE_ENDPOINT, "v1")
)
TEMPL_VMS_COLLECTION_ENDPOINT = utils.lastslash(
    parse.urljoin(TEMPL_V1_COLLECTION_ENDPOINT, "vms")
)
TEMPL_VMSNOPROCESSFILTERS_COLLECTION_ENDPOINT = utils.lastslash(
    parse.urljoin(TEMPL_V1_COLLECTION_ENDPOINT, "vmsnoproccessfilters")
)
TEMPL_NOT_IMPLEMENTED_METHODS_COLLECTION_ENDPOINT = utils.lastslash(
    parse.urljoin(TEMPL_V1_COLLECTION_ENDPOINT, "notimplementedmethods")
)
TEMPL_VM_RESOURCE_ENDPOINT = parse.urljoin(TEMPL_VMS_COLLECTION_ENDPOINT, "%s")
TEMPL_VMS_COLLECTION_ENDPOINT_WITH_FILTER = parse.urljoin(
    TEMPL_VMS_COLLECTION_ENDPOINT, "?%s=%s"
)
TEMPL_POWERON_ACTION_ENDPOINT = parse.urljoin(
    utils.lastslash(TEMPL_VM_RESOURCE_ENDPOINT), "actions/poweron/invoke"
)
TEMPL_POWEROFF_ACTION_ENDPOINT = parse.urljoin(
    utils.lastslash(TEMPL_VM_RESOURCE_ENDPOINT), "actions/poweroff/invoke"
)
TEMPL_POWER_ACTION_ENDPOINT = parse.urljoin(
    utils.lastslash(TEMPL_VM_RESOURCE_ENDPOINT), "actions/power/invoke"
)
TEMPL_POWER_STATE_ACTION_ENDPOINT = parse.urljoin(
    utils.lastslash(TEMPL_VM_RESOURCE_ENDPOINT), "actions/power_state"
)
TEMPL_PORTS_COLLECTION_ENDPOINT = utils.lastslash(
    parse.urljoin(utils.lastslash(TEMPL_VM_RESOURCE_ENDPOINT), "ports")
)
TEMPL_PORTSNONE_COLLECTION_ENDPOINT = utils.lastslash(
    parse.urljoin(utils.lastslash(TEMPL_VM_RESOURCE_ENDPOINT), "none_ports")
)
TEMPL_PORT_RESOURCE_ENDPOINT = parse.urljoin(TEMPL_PORTS_COLLECTION_ENDPOINT, "%s")
TEMPL_PORTNONE_RESOURCE_ENDPOINT = parse.urljoin(
    TEMPL_PORTSNONE_COLLECTION_ENDPOINT, "%s"
)
TEMPL_TAGS_COLLECTION_ENDPOINT = utils.lastslash(
    parse.urljoin(utils.lastslash(TEMPL_VM_RESOURCE_ENDPOINT), "tags")
)
TEMPL_TAG_RESOURCE_ENDPOINT = parse.urljoin(TEMPL_TAGS_COLLECTION_ENDPOINT, "%s")

UUID1 = pyuuid.UUID("00000000-0000-0000-0000-000000000001")
UUID2 = pyuuid.UUID("00000000-0000-0000-0000-000000000002")
UUID3 = pyuuid.UUID("00000000-0000-0000-0000-000000000003")
UUID4 = pyuuid.UUID("00000000-0000-0000-0000-000000000004")
UUID5 = pyuuid.UUID("00000000-0000-0000-0000-000000000005")
ALL_UUIDS = UUID1, UUID2, UUID3, UUID4, UUID5

BAD_UUID = "bad_uuid"

DEADLOCK_EXC = dialect_exc.DeadLock(code=1213, message="fake deadlock")
DEADLOCK_RESPONSE = {
    "type": "DeadLock",
    "message": "Deadlock found when trying to get lock. Original message: "
    "fake deadlock",
    "code": 500,
}


class BaseResourceTestCase(base.BaseWithDbMigrationsTestCase):
    __LAST_MIGRATION__ = "0001-rest-service-tables-migration-e31a12"
    __FIRST_MIGRATION__ = "0001-rest-service-tables-migration-e31a12"

    def get_endpoint(self, template, *args):
        return template % ((self.service_port,) + tuple(args))

    def find_free_port(self):
        with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.bind(("127.0.0.1", 0))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            return s.getsockname()[1]

    def setUp(self):
        super(BaseResourceTestCase, self).setUp()

        self.service_port = self.find_free_port()
        url = parse.urlparse(self.get_endpoint(TEMPL_SERVICE_ENDPOINT))
        self._service = service.RESTService(
            bind_host=url.hostname,
            bind_port=url.port,
            app_root=service.build_wsgi_application(app_root=routes.Root),
        )
        self._service.start()

    def tearDown(self):
        super(BaseResourceTestCase, self).tearDown()

        self._service.stop()


class TestRootResourceTestCase(BaseResourceTestCase):
    def test_get_versions_list(self):

        response = requests.get(self.get_endpoint(TEMPL_ROOT_COLLECTION_ENDPOINT))

        self.assertEqual(response.status_code, 200)
        self.assertEqual(sorted(response.json()), ["specifications", "v1"])


class TestOpenApiSpecificationTestCase(BaseResourceTestCase):
    def test_generate_openapi_specification(self):
        info = {
            "title": "REST API Microservice",
            "description": "REST API Microservice for tests",
            "termsOfService": "https://functional.tests/terms/",
            "contact": {
                "name": "Functional Tests",
                "url": "https://functional.tests/",
                "email": "functional@tests.local",
            },
            "license": {
                "name": "Apache 2.0",
                "url": "https://www.apache.org/licenses/LICENSE-2.0.html",
            },
            "version": "1.2.3",
        }

        response = requests.get(
            self.get_endpoint(TEMPL_OPENAPI_SPEC_ENDPOINT),
        )

        self.assertEqual(200, response.status_code)

        res = response.json()
        # NOTE(v.burygin): to save openapi.yaml
        # import yaml
        # with open("microservice/openapi.yaml", "w") as f:
        #     yaml.safe_dump(res, f, encoding='utf-8', allow_unicode=True)
        self.assertEqual(res["openapi"], "3.0.3")
        self.assertEqual(res["info"], info)


class TestVersionsResourceTestCase(BaseResourceTestCase):
    def test_get_resources_list(self):

        response = requests.get(self.get_endpoint(TEMPL_V1_COLLECTION_ENDPOINT))

        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            [
                "notimplementedmethods",
                "vms",
                "vmsdefsort",
                "vmsnoproccessfilters",
                "vmsnosort",
            ],
        )


class TestVMResourceTestCase(BaseResourceTestCase):
    def tearDown(self):
        super(TestVMResourceTestCase, self).tearDown()
        packers.set_packer(constants.CONTENT_TYPE_APPLICATION_JSON, packers.JSONPacker)

    def _insert_vm_to_db(self, uuid, name, state):
        vm = models.VM(uuid=uuid, name=name, state=state)
        vm.save()
        return vm

    def _vm_exists_in_db(self, uuid):
        try:
            models.VM.objects.get_one(filters={"uuid": uuid})
            return True
        except exceptions.RecordNotFound:
            return False

    def test_not_implemented_error(self):
        endpoint = self.get_endpoint(TEMPL_NOT_IMPLEMENTED_METHODS_COLLECTION_ENDPOINT)
        response = requests.post(endpoint, json={})

        self.assertEqual(response.status_code, 501)
        resp = response.json()
        # Not implemented (method create in NotImplementedMethodsController).
        self.assertIn("create", resp["message"])

        response = requests.get(endpoint)

        self.assertEqual(response.status_code, 501)
        resp = response.json()
        self.assertEqual(resp["message"], "Not implemented.")

        uuid_endpoint = parse.urljoin(endpoint, str(UUID1))

        response = requests.put(uuid_endpoint, json={})

        self.assertEqual(response.status_code, 405)
        resp = response.json()

        # Not allowed (method update in NotImplementedMethodsController).
        self.assertEqual("HTTP method 'UPDATE' is not supported.", resp["message"])

        response = requests.delete(uuid_endpoint)

        self.assertEqual(response.status_code, 501)
        resp = response.json()

        # Not implemented (method delete in NotImplementedMethodsController).
        self.assertIn("delete", resp["message"])

        action_name = "not_implemented_action"
        action_path = "actions/%s/invoke" % action_name
        action_endpoint = parse.urljoin(utils.lastslash(uuid_endpoint), action_path)
        response = requests.post(action_endpoint, json={})

        self.assertEqual(response.status_code, 500)

        resp = response.json()
        # 'NotImplementedMethodsController' object has
        # no attribute 'not_implemented_action'
        self.assertIn(action_name, resp["message"])

    @mock.patch("uuid.uuid4")
    def test_create_vm_resource_successful(self, uuid4_mock):
        RESOURCE_ID = UUID1
        uuid4_mock.return_value = RESOURCE_ID
        vm_request_body = {"name": "test"}
        vm_response_body = {
            "uuid": str(RESOURCE_ID),
            "name": "test",
            "state": "off",
            "status": "active",
            "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
        }
        LOCATION = self.get_endpoint(TEMPL_VM_RESOURCE_ENDPOINT, RESOURCE_ID)

        response = requests.post(
            self.get_endpoint(TEMPL_VMS_COLLECTION_ENDPOINT),
            json=vm_request_body,
        )

        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.headers["location"], LOCATION)
        self.assertEqual(response.json(), vm_response_body)

    def test_get_vm_resource_by_uuid_successful(self):
        RESOURCE_ID = UUID1
        self._insert_vm_to_db(uuid=RESOURCE_ID, name="test", state="off")
        vm_response_body = {
            "uuid": str(RESOURCE_ID),
            "name": "test",
            "state": "off",
            "status": "active",
            "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
        }
        VM_RES_ENDPOINT = self.get_endpoint(TEMPL_VM_RESOURCE_ENDPOINT, RESOURCE_ID)

        response = requests.get(VM_RES_ENDPOINT)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), vm_response_body)

    def test_get_vm_resource_with_null_fields_by_uuid_successful(self):
        RESOURCE_ID = UUID1
        self._insert_vm_to_db(uuid=RESOURCE_ID, name="test", state="off")
        vm_response_body = {
            "uuid": str(RESOURCE_ID),
            "name": "test",
            "state": "off",
            "just-none": None,
            "status": "active",
            "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
        }
        VM_RES_ENDPOINT = self.get_endpoint(TEMPL_VM_RESOURCE_ENDPOINT, RESOURCE_ID)
        packers.set_packer(
            constants.CONTENT_TYPE_APPLICATION_JSON,
            packers.JSONPackerIncludeNullFields,
        )

        response = requests.get(VM_RES_ENDPOINT)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), vm_response_body)

    def test_update_vm_resource_successful(self):
        RESOURCE_ID = UUID1
        self._insert_vm_to_db(uuid=RESOURCE_ID, name="old", state="off")
        vm_request_body = {"name": "new"}
        vm_response_body = {
            "uuid": str(RESOURCE_ID),
            "name": "new",
            "state": "off",
            "status": "active",
            "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
        }
        VM_RES_ENDPOINT = self.get_endpoint(TEMPL_VM_RESOURCE_ENDPOINT, RESOURCE_ID)

        response = requests.put(VM_RES_ENDPOINT, json=vm_request_body)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), vm_response_body)

    def test_update_vm_resource_uuid_in_body(self):
        RESOURCE_ID = UUID1
        self._insert_vm_to_db(uuid=RESOURCE_ID, name="old", state="off")
        vm_request_body = {"uuid": str(UUID2), "name": "new"}

        VM_RES_ENDPOINT = self.get_endpoint(TEMPL_VM_RESOURCE_ENDPOINT, RESOURCE_ID)

        response = requests.put(VM_RES_ENDPOINT, json=vm_request_body)

        self.assertEqual(response.status_code, 400)
        message = response.json()["message"]
        expected_message = (
            "Uuid (%s) in body is not equal to parsed id (%s) from url."
        ) % (UUID2, RESOURCE_ID)
        self.assertEqual(message, expected_message)

        vm_request_body = {"uuid": str(RESOURCE_ID), "name": "new"}
        vm_response_body = {
            "uuid": str(RESOURCE_ID),
            "name": "new",
            "state": "off",
            "status": "active",
            "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
        }
        response = requests.put(VM_RES_ENDPOINT, json=vm_request_body)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), vm_response_body)

    def test_delete_vm_resource_successful(self):
        RESOURCE_ID = UUID1
        self._insert_vm_to_db(uuid=RESOURCE_ID, name="test", state="off")

        VM_RES_ENDPOINT = self.get_endpoint(TEMPL_VM_RESOURCE_ENDPOINT, RESOURCE_ID)

        response = requests.delete(VM_RES_ENDPOINT)

        self.assertEqual(response.status_code, 204)
        self.assertFalse(self._vm_exists_in_db(RESOURCE_ID))

    def test_process_vm_action_successful(self):
        RESOURCE_ID = UUID1
        self._insert_vm_to_db(uuid=RESOURCE_ID, name="test", state="off")
        vm_response_body = {
            "uuid": str(RESOURCE_ID),
            "name": "test",
            "state": "on",
            "status": "active",
            "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
        }
        POWERON_ACT_ENDPOINT = self.get_endpoint(
            TEMPL_POWERON_ACTION_ENDPOINT, RESOURCE_ID
        )

        response = requests.post(POWERON_ACT_ENDPOINT)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), vm_response_body)

    def test_get_collection_vms_successful(self):
        RESOURCE_ID1 = UUID1
        RESOURCE_ID2 = UUID2
        self._insert_vm_to_db(uuid=RESOURCE_ID1, name="test1", state="off")
        self._insert_vm_to_db(uuid=RESOURCE_ID2, name="test2", state="on")
        vm_response_body = [
            {
                "uuid": str(RESOURCE_ID1),
                "name": "test1",
                "state": "off",
                "status": "active",
                "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
                "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            },
            {
                "uuid": str(RESOURCE_ID2),
                "name": "test2",
                "state": "on",
                "status": "active",
                "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
                "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            },
        ]

        response = requests.get(self.get_endpoint(TEMPL_VMS_COLLECTION_ENDPOINT))

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), vm_response_body)

    def test_get_collection_vms_with_field_definition_successful(self):
        RESOURCE_ID1 = UUID1
        RESOURCE_ID2 = UUID2
        self._insert_vm_to_db(uuid=RESOURCE_ID1, name="test1", state="off")
        self._insert_vm_to_db(uuid=RESOURCE_ID2, name="test2", state="on")
        vm_response_body = [
            {
                "uuid": str(RESOURCE_ID1),
            },
            {
                "uuid": str(RESOURCE_ID2),
            },
        ]

        response = requests.get(
            self.get_endpoint(TEMPL_VMS_COLLECTION_ENDPOINT + "?fields=uuid")
        )

        self.assertEqual(200, response.status_code)
        self.assertEqual(vm_response_body, response.json())

    def test_get_collection_vms_with_fields_definition_successful(self):
        RESOURCE_ID1 = UUID1
        RESOURCE_ID2 = UUID2
        self._insert_vm_to_db(uuid=RESOURCE_ID1, name="test1", state="off")
        self._insert_vm_to_db(uuid=RESOURCE_ID2, name="test2", state="on")
        vm_response_body = [
            {
                "uuid": str(RESOURCE_ID1),
                "name": "test1",
            },
            {
                "uuid": str(RESOURCE_ID2),
                "name": "test2",
            },
        ]

        response = requests.get(
            self.get_endpoint(
                TEMPL_VMS_COLLECTION_ENDPOINT + "?fields=uuid&fields=name"
            )
        )

        self.assertEqual(200, response.status_code)
        self.assertEqual(vm_response_body, response.json())

    def test_get_collection_vms_with_filter_by_uuid(self):
        RESOURCE_ID1 = UUID1
        RESOURCE_ID2 = UUID2
        RESOURCE_ID3 = UUID3
        self._insert_vm_to_db(uuid=RESOURCE_ID1, name="test1", state="off")
        self._insert_vm_to_db(uuid=RESOURCE_ID2, name="test2", state="on")
        self._insert_vm_to_db(uuid=RESOURCE_ID3, name="test3", state="off")
        vm_response_body = [
            {
                "uuid": str(RESOURCE_ID2),
                "name": "test2",
                "state": "on",
                "status": "active",
                "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
                "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            }
        ]

        response = requests.get(
            self.get_endpoint(
                "%s?uuid=%s" % (TEMPL_VMS_COLLECTION_ENDPOINT, str(RESOURCE_ID2))
            )
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), vm_response_body)

    def test_get_collection_vms_with_fields_definition_and_filter_by_uuid(
        self,
    ):
        RESOURCE_ID1 = UUID1
        RESOURCE_ID2 = UUID2
        RESOURCE_ID3 = UUID3
        self._insert_vm_to_db(uuid=RESOURCE_ID1, name="test1", state="off")
        self._insert_vm_to_db(uuid=RESOURCE_ID2, name="test2", state="on")
        self._insert_vm_to_db(uuid=RESOURCE_ID3, name="test3", state="off")
        vm_response_body = [{"uuid": str(RESOURCE_ID2), "state": "on"}]

        response = requests.get(
            self.get_endpoint(
                "%s?fields=uuid&uuid=%s&fields=state"
                % (TEMPL_VMS_COLLECTION_ENDPOINT, str(RESOURCE_ID2))
            )
        )

        self.assertEqual(200, response.status_code)
        self.assertEqual(vm_response_body, response.json())

    @parameterized.expand(
        [
            TEMPL_VMS_COLLECTION_ENDPOINT,
            TEMPL_VMSNOPROCESSFILTERS_COLLECTION_ENDPOINT,
        ],
        name_func=make_test_name,
    )
    def test_get_collection_vms_with_filter_by_two_uuid(self, base_url):
        RESOURCE_ID1 = UUID1
        RESOURCE_ID2 = UUID2
        RESOURCE_ID3 = UUID3
        self._insert_vm_to_db(uuid=RESOURCE_ID1, name="test1", state="off")
        self._insert_vm_to_db(uuid=RESOURCE_ID2, name="test2", state="on")
        self._insert_vm_to_db(uuid=RESOURCE_ID3, name="test3", state="off")
        vm_response_body = [
            {
                "uuid": str(RESOURCE_ID1),
                "name": "test1",
                "state": "off",
                "status": "active",
                "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
                "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            },
            {
                "uuid": str(RESOURCE_ID3),
                "name": "test3",
                "state": "off",
                "status": "active",
                "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
                "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            },
        ]

        response = requests.get(
            self.get_endpoint(
                "%s?uuid=%s&uuid=%s"
                % (
                    base_url,
                    str(RESOURCE_ID1),
                    str(RESOURCE_ID3),
                )
            )
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), vm_response_body)

    def test_get_collection_vms_with_sort_one_key(self):
        RESOURCE_ID1 = UUID1
        RESOURCE_ID2 = UUID2
        RESOURCE_ID3 = UUID3
        self._insert_vm_to_db(uuid=RESOURCE_ID1, name="test1", state="off")
        self._insert_vm_to_db(uuid=RESOURCE_ID2, name="test2", state="on")
        self._insert_vm_to_db(uuid=RESOURCE_ID3, name="test3", state="off")
        vm_response_body = [
            {
                "uuid": str(RESOURCE_ID3),
                "name": "test3",
                "state": "off",
                "status": "active",
                "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
                "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            },
            {
                "uuid": str(RESOURCE_ID2),
                "name": "test2",
                "state": "on",
                "status": "active",
                "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
                "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            },
            {
                "uuid": str(RESOURCE_ID1),
                "name": "test1",
                "state": "off",
                "status": "active",
                "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
                "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            },
        ]

        response = requests.get(
            self.get_endpoint(
                "%s?sort_key=name&sort_dir=desc" % (TEMPL_VMS_COLLECTION_ENDPOINT,)
            )
        )

        self.assertEqual(response.status_code, 200)
        for i, item in enumerate(response.json()):
            self.assertEqual(item, vm_response_body[i])

    def test_get_collection_vms_with_sort_one_key_restricted_field(self):
        RESOURCE_ID1 = UUID1
        RESOURCE_ID2 = UUID2
        RESOURCE_ID3 = UUID3
        self._insert_vm_to_db(uuid=RESOURCE_ID1, name="test2", state="on")
        self._insert_vm_to_db(uuid=RESOURCE_ID2, name="test1", state="off")
        self._insert_vm_to_db(uuid=RESOURCE_ID3, name="test3", state="off")
        vm_response_body = [
            {
                "uuid": str(RESOURCE_ID1),
                "name": "test2",
                "state": "on",
                "status": "active",
                "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
                "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            },
            {
                "uuid": str(RESOURCE_ID2),
                "name": "test1",
                "state": "off",
                "status": "active",
                "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
                "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            },
            {
                "uuid": str(RESOURCE_ID3),
                "name": "test3",
                "state": "off",
                "status": "active",
                "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
                "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            },
        ]

        response = requests.get(
            self.get_endpoint(
                "%s?sort_key=name"
                % (
                    utils.lastslash(
                        parse.urljoin(TEMPL_V1_COLLECTION_ENDPOINT, "vmsnosort")
                    ),
                )
            )
        )

        self.assertEqual(response.status_code, 200)
        for i, item in enumerate(response.json()):
            self.assertEqual(item, vm_response_body[i])

    def test_get_collection_vms_with_sort_default(self):
        class Model(models.VM):
            __defaul_sort__ = {"name": "desc"}

        RESOURCE_ID1 = UUID1
        RESOURCE_ID2 = UUID2
        RESOURCE_ID3 = UUID3
        self._insert_vm_to_db(uuid=RESOURCE_ID1, name="test1", state="off")
        self._insert_vm_to_db(uuid=RESOURCE_ID2, name="test2", state="on")
        self._insert_vm_to_db(uuid=RESOURCE_ID3, name="test3", state="off")
        vm_response_body = [
            {
                "uuid": str(RESOURCE_ID3),
                "name": "test3",
                "state": "off",
                "status": "active",
                "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
                "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            },
            {
                "uuid": str(RESOURCE_ID2),
                "name": "test2",
                "state": "on",
                "status": "active",
                "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
                "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            },
            {
                "uuid": str(RESOURCE_ID1),
                "name": "test1",
                "state": "off",
                "status": "active",
                "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
                "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            },
        ]

        response = requests.get(
            self.get_endpoint(
                "%s"
                % (
                    utils.lastslash(
                        parse.urljoin(TEMPL_V1_COLLECTION_ENDPOINT, "vmsdefsort")
                    ),
                )
            )
        )

        self.assertEqual(response.status_code, 200)
        for i, item in enumerate(response.json()):
            self.assertEqual(item, vm_response_body[i])

    def test_get_collection_non_paginated(self):
        RESOURCE_ID1 = UUID1
        RESOURCE_ID2 = UUID2
        self._insert_vm_to_db(uuid=RESOURCE_ID1, name="test1", state="off")
        self._insert_vm_to_db(uuid=RESOURCE_ID2, name="test2", state="on")
        vm_response_body = [
            {
                "uuid": str(RESOURCE_ID1),
                "name": "test1",
                "state": "off",
                "status": "active",
                "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
                "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            },
            {
                "uuid": str(RESOURCE_ID2),
                "name": "test2",
                "state": "on",
                "status": "active",
                "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
                "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            },
        ]

        response = requests.get(self.get_endpoint(TEMPL_VMS_COLLECTION_ENDPOINT))

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), vm_response_body)
        self.assertTrue("X-Pagination-Limit" not in response.headers)
        self.assertTrue("X-Pagination-Marker" not in response.headers)

    @parameterized.expand(
        [
            (None, "all", False),
            (0, "all", False),
            (1, 1, True),
            (3, 3, True),
            (999, "all", False),
        ],
        name_func=make_test_name,
    )
    def test_get_collection_paginated(self, page_limit, expected_len, assert_headers):
        if expected_len == "all":
            expected_len = len(ALL_UUIDS)

        for idx, uuid in enumerate(ALL_UUIDS):
            self._insert_vm_to_db(uuid=uuid, name=f"test{idx}", state="on")

        response = requests.get(
            self.get_endpoint(TEMPL_VMS_COLLECTION_ENDPOINT),
            params={"page_limit": page_limit},
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), expected_len)
        if assert_headers:
            self.assertEqual(response.headers["X-Pagination-Limit"], str(page_limit))
            last_uuid = ALL_UUIDS[expected_len - 1]
            self.assertEqual(response.headers["X-Pagination-Marker"], str(last_uuid))

    @parameterized.expand(
        [
            (page_limit, sort_key, sort_dir)
            for page_limit in [0, 1, 3, 999, None]
            for sort_key in ["uuid", "name", "state", "created"]
            for sort_dir in [None, "asc", "desc"]
        ],
        name_func=make_test_name,
    )
    def test_get_collection_paginated_custom_sort(self, page_limit, sort_key, sort_dir):
        # Create a bunch of docs
        amount = 10
        records_created = [
            self._insert_vm_to_db(
                uuid=pyuuid.uuid4(),
                name=f"test_{i}",
                state="on" if i < 5 else "off",  # non-unique column to sort
            )
            for i in range(amount)
        ]
        # secondary sort by UUID:
        records_created.sort(key=lambda i: i.uuid)
        # primary sort:
        records_created.sort(
            key=lambda i: i.get(sort_key) or i.uuid,
            reverse=sort_dir == "desc",
        )

        # Get pages
        records_fetched = []
        marker = None
        # Loop for overkill amount of iterations
        # Expect "break", or else this test fails
        for _ in range(amount * 2):
            query_params = {
                "page_limit": page_limit,
                "page_marker": marker,
                "sort_key": sort_key,
                "sort_dir": sort_dir,
            }
            response = requests.get(
                self.get_endpoint(TEMPL_VMS_COLLECTION_ENDPOINT),
                params=query_params,
            )
            res_data = response.json()
            marker = res_data[-1]["uuid"] if res_data else None
            self.assertEqual(response.status_code, 200)
            self.assertEqual(
                response.headers.get("X-Pagination-Limit"),
                str(page_limit) if page_limit else None,
            )
            self.assertLessEqual(len(res_data), page_limit or amount)
            for record in res_data:
                # Check that all newly fetched records are not seen before
                self.assertNotIn(record["uuid"], [i["uuid"] for i in records_fetched])
                records_fetched.append(record)
            if len(records_fetched) == amount:
                break
            else:
                self.assertEqual(response.headers.get("X-Pagination-Marker"), marker)
        else:
            self.assertTrue(False, "Did not exit pagination loop properly.")

        self.assertEqual(len(records_fetched), amount)
        self.assertEqual(
            [i["uuid"] for i in records_fetched],
            [str(i.uuid) for i in records_created],
            "Sorting is invalid.",
        )

    def test_get_collection_paginated_with_marker(self):
        RESOURCE_ID1 = UUID1
        RESOURCE_ID2 = UUID2
        RESOURCE_ID3 = UUID3
        self._insert_vm_to_db(uuid=RESOURCE_ID1, name="test1", state="off")
        self._insert_vm_to_db(uuid=RESOURCE_ID2, name="test2", state="on")
        self._insert_vm_to_db(uuid=RESOURCE_ID3, name="test3", state="on")
        vm_response_body = [
            {
                "uuid": str(RESOURCE_ID2),
                "name": "test2",
                "state": "on",
                "status": "active",
                "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
                "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            },
            {
                "uuid": str(RESOURCE_ID3),
                "name": "test3",
                "state": "on",
                "status": "active",
                "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
                "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            },
        ]

        response = requests.get(
            self.get_endpoint(TEMPL_VMS_COLLECTION_ENDPOINT),
            params={
                "page_limit": 2,
                "page_marker": str(RESOURCE_ID1),
            },
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), vm_response_body)
        self.assertEqual(response.headers["X-Pagination-Limit"], "2")
        self.assertEqual(response.headers["X-Pagination-Marker"], str(RESOURCE_ID3))

    def test_get_collection_paginated_with_marker_last(self):
        RESOURCE_ID1 = UUID1
        RESOURCE_ID2 = UUID2
        RESOURCE_ID3 = UUID3
        self._insert_vm_to_db(uuid=RESOURCE_ID1, name="test1", state="off")
        self._insert_vm_to_db(uuid=RESOURCE_ID2, name="test2", state="on")
        self._insert_vm_to_db(uuid=RESOURCE_ID3, name="test3", state="on")
        vm_response_body = [
            {
                "uuid": str(RESOURCE_ID2),
                "name": "test2",
                "state": "on",
                "status": "active",
                "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
                "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            },
            {
                "uuid": str(RESOURCE_ID3),
                "name": "test3",
                "state": "on",
                "status": "active",
                "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
                "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            },
        ]

        response = requests.get(
            self.get_endpoint(TEMPL_VMS_COLLECTION_ENDPOINT)
            + "?page_limit=3&page_marker="
            + str(RESOURCE_ID1),
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), vm_response_body)
        self.assertEqual(response.headers["X-Pagination-Limit"], "3")
        self.assertTrue("X-Pagination-Marker" not in response.headers)

    def test_get_collection_paginated_with_filter(self):
        RESOURCE_ID1 = UUID1
        RESOURCE_ID2 = UUID2
        RESOURCE_ID3 = UUID3
        self._insert_vm_to_db(uuid=RESOURCE_ID1, name="test1", state="off")
        self._insert_vm_to_db(uuid=RESOURCE_ID2, name="test2", state="on")
        self._insert_vm_to_db(uuid=RESOURCE_ID3, name="test3", state="on")
        vm_response_body = [
            {
                "uuid": str(RESOURCE_ID2),
                "name": "test2",
                "state": "on",
                "status": "active",
                "created": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
                "updated": types.DEFAULT_DATE.strftime(types.OPENAPI_DATETIME_FMT),
            }
        ]

        response = requests.get(
            self.get_endpoint(
                "%s?state=on" % (TEMPL_VMS_COLLECTION_ENDPOINT,) + "&page_limit=1"
            ),
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), vm_response_body)
        self.assertEqual(response.headers["X-Pagination-Limit"], "1")
        self.assertEqual(response.headers["X-Pagination-Marker"], str(RESOURCE_ID2))

    def test_get_collection_paginated_limit_nonnumeric_negative(self):
        RESOURCE_ID1 = UUID1
        RESOURCE_ID2 = UUID2
        self._insert_vm_to_db(uuid=RESOURCE_ID1, name="test1", state="off")
        self._insert_vm_to_db(uuid=RESOURCE_ID2, name="test2", state="on")

        response = requests.get(
            self.get_endpoint(TEMPL_VMS_COLLECTION_ENDPOINT) + "?page_limit=a"
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json()["type"], "ParseError")

    def test_get_collection_paginated_limit_number_negative(self):
        RESOURCE_ID1 = UUID1
        RESOURCE_ID2 = UUID2
        self._insert_vm_to_db(uuid=RESOURCE_ID1, name="test1", state="off")
        self._insert_vm_to_db(uuid=RESOURCE_ID2, name="test2", state="on")

        response = requests.get(
            self.get_endpoint(TEMPL_VMS_COLLECTION_ENDPOINT) + "?page_limit=-1"
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json()["type"], "ParseError")

    def test_get_collection_paginated_marker_negative(self):
        RESOURCE_ID1 = UUID1
        RESOURCE_ID2 = UUID2
        self._insert_vm_to_db(uuid=RESOURCE_ID1, name="test1", state="off")
        self._insert_vm_to_db(uuid=RESOURCE_ID2, name="test2", state="on")

        response = requests.get(
            self.get_endpoint(TEMPL_VMS_COLLECTION_ENDPOINT)
            + "?page_limit=-1&page_marker=a"
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json()["type"], "ParseError")


class TestNestedResourceTestCase(BaseResourceTestCase):
    __LAST_MIGRATION__ = "0002-0-rest-service-data-for-test-nested-resource-c17a60"

    def setUp(self):
        super(TestNestedResourceTestCase, self).setUp()

        self.vm1 = models.VM.objects.get_one(filters={"uuid": filters.EQ(UUID1)})
        self.vm2 = models.VM.objects.get_one(filters={"uuid": filters.EQ(UUID2)})

    def tearDown(self):
        super(TestNestedResourceTestCase, self).tearDown()

    @mock.patch("uuid.uuid4")
    def test_create_nested_resource_successful(self, uuid4_mock):
        VM_RESOURCE_ID = UUID1
        PORT_RESOURCE_ID = UUID3
        uuid4_mock.return_value = PORT_RESOURCE_ID
        port_request_body = {"mac": "00:00:00:00:00:03"}
        port_response_body = {
            "uuid": str(PORT_RESOURCE_ID),
            "mac": "00:00:00:00:00:03",
            "vm": parse.urlparse(
                self.get_endpoint(TEMPL_VM_RESOURCE_ENDPOINT, VM_RESOURCE_ID)
            ).path,
            "some-field2": "some_field2",
            "some-field3": "some_field3",
            "some-field4": "some_field4",
        }
        LOCATION = self.get_endpoint(
            TEMPL_PORT_RESOURCE_ENDPOINT, VM_RESOURCE_ID, PORT_RESOURCE_ID
        )

        response = requests.post(
            self.get_endpoint(TEMPL_PORTS_COLLECTION_ENDPOINT, VM_RESOURCE_ID),
            json=port_request_body,
        )

        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.headers["location"], LOCATION)
        self.assertEqual(port_response_body, response.json())

    def test_update_nested_resource_successful(self):
        VM_RESOURCE_ID = UUID1
        PORT_RESOURCE_ID = UUID3
        port = models.Port(uuid=PORT_RESOURCE_ID, mac="00:00:00:00:00:03", vm=self.vm1)
        port.save()
        port_request_body = {"mac": "00:00:00:00:00:04"}
        port_response_body = {
            "uuid": str(PORT_RESOURCE_ID),
            "mac": "00:00:00:00:00:04",
            "vm": parse.urlparse(
                self.get_endpoint(TEMPL_VM_RESOURCE_ENDPOINT, VM_RESOURCE_ID)
            ).path,
            "some-field1": "some_field1",
            "some-field2": "some_field2",
            "some-field3": "some_field3",
        }

        response = requests.put(
            self.get_endpoint(
                TEMPL_PORT_RESOURCE_ENDPOINT, VM_RESOURCE_ID, PORT_RESOURCE_ID
            ),
            json=port_request_body,
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(port_response_body, response.json())

    def test_get_nested_resource_successful(self):
        VM_RESOURCE_ID = UUID1
        PORT_RESOURCE_ID = UUID3
        port = models.Port(uuid=PORT_RESOURCE_ID, mac="00:00:00:00:00:03", vm=self.vm1)
        port.save()
        port_response_body = {
            "uuid": str(PORT_RESOURCE_ID),
            "mac": "00:00:00:00:00:03",
            "vm": parse.urlparse(
                self.get_endpoint(TEMPL_VM_RESOURCE_ENDPOINT, VM_RESOURCE_ID)
            ).path,
            "some-field1": "some_field1",
            "some-field2": "some_field2",
            "some-field4": "some_field4",
        }

        response = requests.get(
            self.get_endpoint(
                TEMPL_PORT_RESOURCE_ENDPOINT, VM_RESOURCE_ID, PORT_RESOURCE_ID
            )
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(port_response_body, response.json())

    def test_get_nested_resource_none_successful(self):
        VM_RESOURCE_ID = UUID1
        PORT_RESOURCE_ID = UUID3
        port = models.Port(uuid=PORT_RESOURCE_ID, mac="00:00:00:00:00:03", vm=self.vm1)
        port.save()
        port_response_body = {
            "uuid": str(PORT_RESOURCE_ID),
            "mac": "00:00:00:00:00:03",
            "vm": parse.urlparse(
                self.get_endpoint(TEMPL_VM_RESOURCE_ENDPOINT, VM_RESOURCE_ID)
            ).path,
            "some-field1": "some_field1",
            "some-field2": "some_field2",
            "some-field4": "some_field4",
            "some-field5": None,
        }

        response = requests.get(
            self.get_endpoint(
                TEMPL_PORTNONE_RESOURCE_ENDPOINT,
                VM_RESOURCE_ID,
                PORT_RESOURCE_ID,
            )
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(port_response_body, response.json())

    def test_get_nested_resource_with_fields_definition_successful(self):
        VM_RESOURCE_ID = UUID1
        PORT_RESOURCE_ID = UUID3
        port = models.Port(uuid=PORT_RESOURCE_ID, mac="00:00:00:00:00:03", vm=self.vm1)
        port.save()
        port_response_body = {
            "uuid": str(PORT_RESOURCE_ID),
            "vm": parse.urlparse(
                self.get_endpoint(TEMPL_VM_RESOURCE_ENDPOINT, VM_RESOURCE_ID)
            ).path,
            "some-field2": "some_field2",
        }

        response = requests.get(
            self.get_endpoint(
                TEMPL_PORT_RESOURCE_ENDPOINT, VM_RESOURCE_ID, PORT_RESOURCE_ID
            )
            + "?fields=uuid&fields=vm&fields=some-field2"
        )

        self.assertEqual(200, response.status_code)
        self.assertEqual(port_response_body, response.json())

    def test_get_ports_collection_successful(self):
        VM_RESOURCE_ID = UUID1
        PORT1_RESOURCE_ID = UUID3
        PORT2_RESOURCE_ID = UUID4
        PORT3_RESOURCE_ID = UUID5
        port1 = models.Port(
            uuid=PORT1_RESOURCE_ID, mac="00:00:00:00:00:03", vm=self.vm1
        )
        port1.save()
        port2 = models.Port(
            uuid=PORT2_RESOURCE_ID, mac="00:00:00:00:00:04", vm=self.vm1
        )
        port2.save()
        port3 = models.Port(
            uuid=PORT3_RESOURCE_ID, mac="00:00:00:00:00:05", vm=self.vm2
        )
        port3.save()
        ports_response_body = [
            {
                "uuid": str(PORT1_RESOURCE_ID),
                "mac": "00:00:00:00:00:03",
                "vm": parse.urlparse(
                    self.get_endpoint(TEMPL_VM_RESOURCE_ENDPOINT, VM_RESOURCE_ID)
                ).path,
                "some-field1": "some_field1",
                "some-field3": "some_field3",
                "some-field4": "some_field4",
                "unique-field": str(PORT1_RESOURCE_ID),
            },
            {
                "uuid": str(PORT2_RESOURCE_ID),
                "mac": "00:00:00:00:00:04",
                "vm": parse.urlparse(
                    self.get_endpoint(TEMPL_VM_RESOURCE_ENDPOINT, VM_RESOURCE_ID)
                ).path,
                "some-field1": "some_field1",
                "some-field3": "some_field3",
                "some-field4": "some_field4",
                "unique-field": str(PORT2_RESOURCE_ID),
            },
        ]

        response = requests.get(
            self.get_endpoint(TEMPL_PORTS_COLLECTION_ENDPOINT, VM_RESOURCE_ID)
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(ports_response_body, response.json())

    def test_delete_nested_resource_successful(self):
        VM_RESOURCE_ID = UUID1
        PORT_RESOURCE_ID = UUID3
        port = models.Port(uuid=PORT_RESOURCE_ID, mac="00:00:00:00:00:03", vm=self.vm1)
        port.save()

        response = requests.delete(
            self.get_endpoint(
                TEMPL_PORT_RESOURCE_ENDPOINT, VM_RESOURCE_ID, PORT_RESOURCE_ID
            )
        )

        self.assertEqual(response.status_code, 204)
        self.assertRaises(
            exceptions.RecordNotFound,
            models.Port.objects.get_one,
            filters={"uuid": PORT_RESOURCE_ID},
        )

    def test_get_ports_collection_pagination_custom_props_iterator(self):
        VM_RESOURCE_ID = UUID1
        PORT1_RESOURCE_ID = UUID3
        PORT2_RESOURCE_ID = UUID4
        PORT3_RESOURCE_ID = UUID5
        port1 = models.Port(
            uuid=PORT1_RESOURCE_ID, mac="00:00:00:00:00:03", vm=self.vm1
        )
        port1.save()
        port2 = models.Port(
            uuid=PORT2_RESOURCE_ID, mac="00:00:00:00:00:04", vm=self.vm1
        )
        port2.save()
        port3 = models.Port(
            uuid=PORT3_RESOURCE_ID, mac="00:00:00:00:00:05", vm=self.vm2
        )
        port3.save()
        ports_response_body = [
            {
                "uuid": str(PORT2_RESOURCE_ID),
                "mac": "00:00:00:00:00:04",
                "vm": parse.urlparse(
                    self.get_endpoint(TEMPL_VM_RESOURCE_ENDPOINT, VM_RESOURCE_ID)
                ).path,
                "some-field1": "some_field1",
                "some-field3": "some_field3",
                "some-field4": "some_field4",
                "unique-field": str(PORT2_RESOURCE_ID),
            }
        ]

        response = requests.get(
            self.get_endpoint(TEMPL_PORTS_COLLECTION_ENDPOINT, VM_RESOURCE_ID)
            + "?unique-field=%s&page_limit=1" % str(PORT2_RESOURCE_ID),
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(ports_response_body, response.json())
        self.assertEqual(response.headers["X-Pagination-Limit"], "1")
        self.assertEqual(
            response.headers["X-Pagination-Marker"], str(PORT2_RESOURCE_ID)
        )


class TestMultipleIdProperties(BaseResourceTestCase):
    __LAST_MIGRATION__ = "0002-0-rest-service-data-for-test-nested-resource-c17a60"

    def setUp(self):
        super(TestMultipleIdProperties, self).setUp()

        self.vm1 = models.VM.objects.get_one(filters={"uuid": filters.EQ(UUID1)})
        self.vm2 = models.VM.objects.get_one(filters={"uuid": filters.EQ(UUID2)})

    def _swap_tag_visibility(self, vm_uuid, tag_uuid, new_visible):
        tag_response_body = {
            "uuid": str(tag_uuid),
            "vm": parse.urlparse(
                self.get_endpoint(TEMPL_VM_RESOURCE_ENDPOINT, vm_uuid)
            ).path,
            "name": "tagname",
            "visible": new_visible,
        }

        response = requests.put(
            self.get_endpoint(TEMPL_TAG_RESOURCE_ENDPOINT, vm_uuid, "tagname"),
            json={"visible": new_visible},
        )
        self.assertEqual(200, response.status_code)
        self.assertEqual(response.json(), tag_response_body)

    def test_update_same_name_uuid(self):
        models.Tag(uuid=UUID3, vm=self.vm1, name="tagname", visible=True).save()
        models.Tag(uuid=UUID4, vm=self.vm2, name="tagname", visible=False).save()

        self._swap_tag_visibility(UUID1, UUID3, False)
        self._swap_tag_visibility(UUID2, UUID4, True)


class ResourceExceptionsTestCase(BaseResourceTestCase):
    def _insert_vm_to_db(self, uuid, name, state):
        vm = models.VM(uuid=uuid, name=name, state=state)
        vm.save()

    def test_create_parse_error_exception(self):
        vm_request_body = {"uuid": BAD_UUID, "name": "test"}

        response = requests.post(
            self.get_endpoint(TEMPL_VMS_COLLECTION_ENDPOINT),
            json=vm_request_body,
        )

        message = response.json()["message"]
        self.assertEqual(response.status_code, 400)
        expected_message = "Can't parse value: %s=%s." % ("uuid", BAD_UUID)
        self.assertEqual(message, expected_message)

    def test_filter_parse_error_exception(self):
        self._insert_vm_to_db(uuid=UUID1, name="test", state="off")

        end_point = self.get_endpoint(
            TEMPL_VMS_COLLECTION_ENDPOINT_WITH_FILTER, "uuid", BAD_UUID
        )

        response = requests.get(end_point)

        message = response.json()["message"]
        self.assertEqual(response.status_code, 400)
        expected_message = "Can't parse value: %s=%s." % ("uuid", BAD_UUID)
        self.assertEqual(message, expected_message)

    def test_resource_id_parse_error_exception(self):
        end_point = self.get_endpoint(TEMPL_VM_RESOURCE_ENDPOINT, BAD_UUID)

        response = requests.get(end_point)

        message = response.json()["message"]
        self.assertEqual(response.status_code, 400)
        expected_message = "Can't parse value: %s=%s." % ("uuid", BAD_UUID)
        self.assertEqual(message, expected_message)


class TestNestedResourceForUnpackerTestCase(BaseResourceTestCase):
    __LAST_MIGRATION__ = "0002-1-rest-service-data-for-test-unpacker-1a9112"

    def test_get_resource_by_uri(self):
        uri = "/v1/vms/%s/ports/%s/ip_addresses/%s" % (
            UUID1,
            UUID2,
            UUID3,
        )
        req = request.Request.blank(uri)

        result = resources.ResourceMap.get_resource(req, uri)

        self.assertEqual(
            models.IpAddress.objects.get_one(filters={"uuid": filters.EQ(UUID3)}),
            result,
        )


def raise_deadlock_exc_once(counter, original_method, obj, *args, **kwargs):
    if counter["attempt"] == 0:
        counter["attempt"] += 1
        raise DEADLOCK_EXC
    else:
        return original_method(obj, *args, **kwargs)


class TestRetryOnErrorMiddlewareBaseResourceTestCase(BaseResourceTestCase):
    @mock.patch(
        "restalchemy.storage.sql.tables.SQLTable.insert",
        side_effect=DEADLOCK_EXC,
    )
    def test_create_vm_when_deadlock_raises(self, model_insert_mock):
        response = requests.post(
            self.get_endpoint(TEMPL_VMS_COLLECTION_ENDPOINT),
            json={"name": "test"},
        )
        self.assertEqual(500, response.status_code)
        self.assertEqual(DEADLOCK_RESPONSE, response.json())
        self.assertEqual(2, model_insert_mock.call_count)
        self.assertEqual(0, len(models.VM.objects.get_all()))

    @mock.patch(
        "restalchemy.storage.sql.tables.SQLTable.insert",
        side_effect=partial(
            raise_deadlock_exc_once, collections.Counter(), SQLTable.insert
        ),
        autospec=True,
    )
    def test_create_vm_when_deadlock_raises_once(self, model_insert_mock):
        response = requests.post(
            self.get_endpoint(TEMPL_VMS_COLLECTION_ENDPOINT),
            json={"name": "test"},
        )
        self.assertEqual(201, response.status_code)
        self.assertEqual(2, model_insert_mock.call_count)
        self.assertIsNotNone(models.VM.objects.get_one(filters={"name": "test"}))

    @mock.patch(
        "restalchemy.storage.sql.tables.SQLTable.update",
        side_effect=DEADLOCK_EXC,
    )
    def test_update_vm_when_deadlock_raises(self, model_update_mock):
        models.VM(uuid=UUID1, name="old", state="off").save()
        response = requests.put(
            self.get_endpoint(TEMPL_VM_RESOURCE_ENDPOINT, UUID1),
            json={"name": "new"},
        )
        self.assertEqual(500, response.status_code)
        self.assertEqual(DEADLOCK_RESPONSE, response.json())
        self.assertEqual(2, model_update_mock.call_count)
        self.assertIsNotNone(
            models.VM.objects.get_one(
                filters={"uuid": UUID1, "name": "old", "state": "off"}
            )
        )

    @mock.patch(
        "restalchemy.storage.sql.tables.SQLTable.update",
        side_effect=partial(
            raise_deadlock_exc_once, collections.Counter(), SQLTable.update
        ),
        autospec=True,
    )
    def test_update_vm_when_deadlock_raises_once(self, model_update_mock):
        models.VM(uuid=UUID1, name="old", state="off").save()
        response = requests.put(
            self.get_endpoint(TEMPL_VM_RESOURCE_ENDPOINT, UUID1),
            json={"name": "new"},
        )
        self.assertEqual(200, response.status_code)
        self.assertEqual(2, model_update_mock.call_count)
        self.assertIsNotNone(
            models.VM.objects.get_one(
                filters={"uuid": UUID1, "name": "new", "state": "off"}
            )
        )

    @mock.patch(
        "restalchemy.storage.sql.tables.SQLTable.delete",
        side_effect=DEADLOCK_EXC,
    )
    def test_delete_vm_when_deadlock_raises(self, model_delete_mock):
        models.VM(uuid=UUID1, name="old", state="off").save()
        response = requests.delete(self.get_endpoint(TEMPL_VM_RESOURCE_ENDPOINT, UUID1))
        self.assertEqual(500, response.status_code)
        self.assertEqual(DEADLOCK_RESPONSE, response.json())
        self.assertEqual(2, model_delete_mock.call_count)
        self.assertIsNotNone(
            models.VM.objects.get_one(
                filters={"uuid": UUID1, "name": "old", "state": "off"}
            )
        )

    @mock.patch(
        "restalchemy.storage.sql.tables.SQLTable.delete",
        side_effect=partial(
            raise_deadlock_exc_once, collections.Counter(), SQLTable.delete
        ),
        autospec=True,
    )
    def test_delete_vm_when_deadlock_raises_once(self, model_delete_mock):
        models.VM(uuid=UUID1, name="old", state="off").save()
        response = requests.delete(self.get_endpoint(TEMPL_VM_RESOURCE_ENDPOINT, UUID1))
        self.assertEqual(204, response.status_code)
        self.assertEqual(2, model_delete_mock.call_count)
        self.assertEqual(0, len(models.VM.objects.get_all()))

    @mock.patch(
        "restalchemy.storage.sql.tables.SQLTable.update",
        side_effect=DEADLOCK_EXC,
    )
    def test_vm_power_on_when_deadlock_raises(self, model_update_mock):
        models.VM(uuid=UUID1, name="old", state="off").save()
        response = requests.post(
            self.get_endpoint(TEMPL_POWERON_ACTION_ENDPOINT, UUID1)
        )
        self.assertEqual(500, response.status_code)
        self.assertEqual(DEADLOCK_RESPONSE, response.json())
        self.assertEqual(2, model_update_mock.call_count)
        self.assertIsNotNone(
            models.VM.objects.get_one(
                filters={"uuid": UUID1, "name": "old", "state": "off"}
            )
        )

    @mock.patch(
        "restalchemy.storage.sql.tables.SQLTable.update",
        side_effect=partial(
            raise_deadlock_exc_once, collections.Counter(), SQLTable.update
        ),
        autospec=True,
    )
    def test_vm_power_on_when_deadlock_raises_once(self, model_update_mock):
        models.VM(uuid=UUID1, name="old", state="off").save()
        response = requests.post(
            self.get_endpoint(TEMPL_POWERON_ACTION_ENDPOINT, UUID1)
        )
        self.assertEqual(200, response.status_code)
        self.assertEqual(2, model_update_mock.call_count)
        self.assertIsNotNone(
            models.VM.objects.get_one(
                filters={"uuid": UUID1, "name": "old", "state": "on"}
            )
        )

    @mock.patch("restalchemy.storage.sql.tables.SQLTable.update")
    def test_vm_power_off_when_deadlock_raises(self, model_update_mock):
        model_update_mock.side_effect = DEADLOCK_EXC
        models.VM(uuid=UUID1, name="old", state="on").save()
        response = requests.post(
            self.get_endpoint(TEMPL_POWEROFF_ACTION_ENDPOINT, UUID1)
        )
        self.assertEqual(500, response.status_code)
        self.assertEqual(DEADLOCK_RESPONSE, response.json())
        self.assertEqual(2, model_update_mock.call_count)
        self.assertIsNotNone(
            models.VM.objects.get_one(
                filters={"uuid": UUID1, "name": "old", "state": "on"}
            )
        )

    @mock.patch(
        "restalchemy.storage.sql.tables.SQLTable.update",
        side_effect=partial(
            raise_deadlock_exc_once, collections.Counter(), SQLTable.update
        ),
        autospec=True,
    )
    def test_vm_power_off_when_deadlock_raises_once(self, model_update_mock):
        models.VM(uuid=UUID1, name="old", state="on").save()
        response = requests.post(
            self.get_endpoint(TEMPL_POWEROFF_ACTION_ENDPOINT, UUID1)
        )
        self.assertEqual(200, response.status_code)
        self.assertEqual(2, model_update_mock.call_count)
        self.assertIsNotNone(
            models.VM.objects.get_one(
                filters={"uuid": UUID1, "name": "old", "state": "off"}
            )
        )

    @mock.patch(
        "restalchemy.storage.sql.tables.SQLTable.update",
        side_effect=partial(
            raise_deadlock_exc_once, collections.Counter(), SQLTable.update
        ),
        autospec=True,
    )
    def test_vm_power_off_with_data(self, model_update_mock):
        models.VM(uuid=UUID1, name="old", state="on").save()
        # requests data
        response = requests.post(
            self.get_endpoint(TEMPL_POWER_ACTION_ENDPOINT, UUID1),
            data={"state": "off", "my_field": "test_value"},
        )
        self.assertEqual(200, response.status_code)
        self.assertIsNotNone(
            models.VM.objects.get_one(
                filters={"uuid": UUID1, "name": "old", "state": "off"}
            )
        )
        self.assertEqual(2, model_update_mock.call_count)

    @mock.patch(
        "restalchemy.storage.sql.tables.SQLTable.update",
        side_effect=partial(
            raise_deadlock_exc_once, collections.Counter(), SQLTable.update
        ),
        autospec=True,
    )
    def test_vm_power_off_with_json(self, model_update_mock):
        models.VM(uuid=UUID1, name="old", state="on").save()
        # requests json (auto header application/json)
        response = requests.post(
            self.get_endpoint(TEMPL_POWER_ACTION_ENDPOINT, UUID1),
            json={"state": "off", "my_field": "test_value"},
        )
        self.assertEqual(200, response.status_code)
        self.assertIsNotNone(
            models.VM.objects.get_one(
                filters={"uuid": UUID1, "name": "old", "state": "off"}
            )
        )
        self.assertEqual(2, model_update_mock.call_count)

    def test_vm_get_power_state(self):
        models.VM(uuid=UUID1, name="old", state="on").save()
        # manual header application/json and no body
        headers = {"Content-type": "application/json"}
        response = requests.get(
            self.get_endpoint(TEMPL_POWER_STATE_ACTION_ENDPOINT, UUID1),
            headers=headers,
            json={"state": "off", "my_field": "test_value"},
        )
        self.assertEqual(200, response.status_code)
        self.assertEqual({"state": "on"}, response.json())

        response = requests.get(
            self.get_endpoint(TEMPL_POWER_STATE_ACTION_ENDPOINT, UUID1),
            data={"state": "off", "my_field": "test_value"},
        )
        self.assertEqual(200, response.status_code)
        self.assertEqual({"state": "on"}, response.json())

    def test_vm_power_off_with_manual_json(self):
        models.VM(uuid=UUID1, name="old", state="on").save()
        # manual header application/json and no body
        headers = {"Content-type": "application/json"}
        response = requests.post(
            self.get_endpoint(TEMPL_POWER_ACTION_ENDPOINT, UUID1),
            headers=headers,
        )
        self.assertEqual(500, response.status_code)


class TestRetryOnErrorMiddlewareNestedResourceTestCase(BaseResourceTestCase):
    __LAST_MIGRATION__ = "0002-0-rest-service-data-for-test-nested-resource-c17a60"

    def setUp(self):
        super(TestRetryOnErrorMiddlewareNestedResourceTestCase, self).setUp()

        self.vm1 = models.VM.objects.get_one(filters={"uuid": filters.EQ(UUID1)})

    @mock.patch(
        "restalchemy.storage.sql.tables.SQLTable.insert",
        side_effect=DEADLOCK_EXC,
    )
    def test_create_port_when_deadlock_raises(self, model_insert_mock):
        response = requests.post(
            self.get_endpoint(TEMPL_PORTS_COLLECTION_ENDPOINT, UUID1),
            json={"mac": "00:00:00:00:00:01"},
        )
        self.assertEqual(500, response.status_code)
        self.assertEqual(DEADLOCK_RESPONSE, response.json())
        self.assertEqual(2, model_insert_mock.call_count)
        self.assertEqual(0, len(models.Port.objects.get_all()))

    @mock.patch(
        "restalchemy.storage.sql.tables.SQLTable.insert",
        side_effect=partial(
            raise_deadlock_exc_once, collections.Counter(), SQLTable.insert
        ),
        autospec=True,
    )
    def test_create_port_when_deadlock_raises_once(self, model_insert_mock):
        response = requests.post(
            self.get_endpoint(TEMPL_PORTS_COLLECTION_ENDPOINT, UUID1),
            json={"mac": "00:00:00:00:00:01"},
        )
        self.assertEqual(201, response.status_code)
        self.assertEqual(2, model_insert_mock.call_count)
        self.assertIsNotNone(
            models.Port.objects.get_one(filters={"mac": "00:00:00:00:00:01"})
        )

    @mock.patch(
        "restalchemy.storage.sql.tables.SQLTable.update",
        side_effect=DEADLOCK_EXC,
    )
    def test_update_port_when_deadlock_raises(self, model_update_mock):
        models.Port(uuid=UUID3, mac="00:00:00:00:00:03", vm=self.vm1).save()
        response = requests.put(
            self.get_endpoint(TEMPL_PORT_RESOURCE_ENDPOINT, UUID1, UUID3),
            json={"mac": "00:00:00:00:00:04"},
        )
        self.assertEqual(500, response.status_code)
        self.assertEqual(DEADLOCK_RESPONSE, response.json())
        self.assertEqual(2, model_update_mock.call_count)
        self.assertIsNotNone(
            models.Port.objects.get_one(
                filters={"uuid": UUID3, "mac": "00:00:00:00:00:03"}
            )
        )

    @mock.patch(
        "restalchemy.storage.sql.tables.SQLTable.update",
        side_effect=partial(
            raise_deadlock_exc_once, collections.Counter(), SQLTable.update
        ),
        autospec=True,
    )
    def test_update_port_when_deadlock_raises_once(self, model_update_mock):
        models.Port(uuid=UUID3, mac="00:00:00:00:00:03", vm=self.vm1).save()
        response = requests.put(
            self.get_endpoint(TEMPL_PORT_RESOURCE_ENDPOINT, UUID1, UUID3),
            json={"mac": "00:00:00:00:00:04"},
        )
        self.assertEqual(200, response.status_code)
        self.assertEqual(2, model_update_mock.call_count)
        self.assertIsNotNone(
            models.Port.objects.get_one(
                filters={"uuid": UUID3, "mac": "00:00:00:00:00:04"}
            )
        )

    @mock.patch(
        "restalchemy.storage.sql.tables.SQLTable.delete",
        side_effect=DEADLOCK_EXC,
    )
    def test_delete_port_when_deadlock_raises(self, model_delete_mock):
        models.Port(uuid=UUID3, mac="00:00:00:00:00:03", vm=self.vm1).save()
        response = requests.delete(
            self.get_endpoint(TEMPL_PORT_RESOURCE_ENDPOINT, UUID1, UUID3)
        )
        self.assertEqual(500, response.status_code)
        self.assertEqual(DEADLOCK_RESPONSE, response.json())
        self.assertEqual(2, model_delete_mock.call_count)
        self.assertIsNotNone(
            models.Port.objects.get_one(
                filters={"uuid": UUID3, "mac": "00:00:00:00:00:03"}
            )
        )

    @mock.patch(
        "restalchemy.storage.sql.tables.SQLTable.delete",
        side_effect=partial(
            raise_deadlock_exc_once, collections.Counter(), SQLTable.delete
        ),
        autospec=True,
    )
    def test_delete_port_when_deadlock_raises_once(self, model_delete_mock):
        models.Port(uuid=UUID3, mac="00:00:00:00:00:03", vm=self.vm1).save()
        response = requests.delete(
            self.get_endpoint(TEMPL_PORT_RESOURCE_ENDPOINT, UUID1, UUID3)
        )
        self.assertEqual(204, response.status_code)
        self.assertEqual(2, model_delete_mock.call_count)
        self.assertEqual(0, len(models.Port.objects.get_all()))
